#!/usr/bin/env python

import collections
import os
import os.path as op
import re
import sys
from argparse import ArgumentParser
from importlib.machinery import SourceFileLoader

import simplejson as json
from jsonschema import ValidationError

import boutiques
from boutiques.logger import raise_error
from boutiques.util.utils import (
    customSortDescriptorByKey,
    customSortInvocationByInput,
    importCatcher,
    loadJson,
)
from boutiques.validator import validate_descriptor


class ImportError(Exception):
    pass


class Importer:

    def __init__(
        self,
        input_descriptor,
        output_descriptor,
        input_invocation,
        output_invocation,
    ):
        self.input_descriptor = input_descriptor
        self.output_descriptor = output_descriptor
        self.input_invocation = input_invocation
        self.output_invocation = output_invocation

    def upgrade_04(self):
        """
         Differences between 0.4 and current (0.5):
           -schema version (obv)
           -singularity should now be represented same as docker
           -walltime should be part of suggested_resources structure

        I.e.
        "schema-version": "0.4",
                    ...... becomes.....
        "schema-version": "0.5",

        I.e.
        "container-image": {
          "type": "singularity",
          "url": "shub://gkiar/ndmg-cbrain:master"
          },
                    ...... becomes.....
        "container-image": {
          "type": "singularity",
          "image": "gkiar/ndmg-cbrain:master",
          "index": "shub://",
        },

        I.e.
        "walltime-estimate": 3600,
                    ...... becomes.....
        "suggested-resources": {
          "walltime-estimate": 3600
        },
        """
        descriptor = loadJson(self.input_descriptor)

        if descriptor["schema-version"] != "0.4":
            raise_error(
                ImportError,
                "The input descriptor must have " "'schema-version'=0.4",
            )
        descriptor["schema-version"] = "0.5"

        if "container-image" in descriptor.keys():
            if "singularity" == descriptor["container-image"]["type"]:
                url = descriptor["container-image"]["url"]
                img = url.split("://")
                if len(img) == 1:
                    descriptor["container-image"]["image"] = img[0]
                elif len(img) == 2:
                    descriptor["container-image"]["image"] = img[1]
                    descriptor["container-image"]["index"] = img[0] + "://"
                del descriptor["container-image"]["url"]
            elif "docker" == descriptor["container-image"]["type"] and descriptor[
                "container-image"
            ].get("index"):
                url = descriptor["container-image"]["index"].split("://")[-1]
                descriptor["container-image"]["index"] = url

        if "walltime-estimate" in descriptor.keys():
            descriptor["suggested-resources"] = {
                "walltime-estimate": descriptor["walltime-estimate"]
            }
            del descriptor["walltime-estimate"]

        with open(self.output_descriptor, "w") as fhandle:
            fhandle.write(json.dumps(customSortDescriptorByKey(descriptor), indent=4))
        validate_descriptor(loadJson(self.output_descriptor))

    def get_entry_point(self, input_descriptor):
        entrypoint = None
        with open(os.path.join(self.input_descriptor, "Dockerfile")) as f:
            content = f.readlines()
        for line in content:
            if line.startswith("ENTRYPOINT"):
                entrypoint_values = line.strip("ENTRYPOINT").strip()

                # According to DOCKER documentation the ENTRYPOINT can
                # be like:
                #   > ENTRYPOINT ["executable", "param1", "param2"]
                # or:
                #   > ENTRYPOINT command param1 param2
                if not entrypoint_values:
                    return None

                if entrypoint_values.startswith("["):
                    # Replacing entrypoint_values
                    # with the Python-interpreted list version
                    entrypoint_values = json.loads(entrypoint_values)
                    # Adding single quotes around list items with spaces therein
                    entrypoint_values = [
                        ev if " " not in ev else f"'{ev}'" for ev in entrypoint_values
                    ]
                    return " ".join(entrypoint_values)

        return entrypoint

    def import_docopt(self):
        # input_descriptor is the .py containing docopt script
        docstring = (
            SourceFileLoader("docopt_script", self.input_descriptor)
            .load_module()
            .__doc__
        )
        # Init docopt importer and start import
        docoptImporter = Docopt_Importer(docstring, self.output_descriptor)

        with open(self.output_descriptor, "w") as output:
            output.write(json.dumps(docoptImporter.descriptor, indent=4))

    def import_bids(self):
        path, fil = os.path.split(__file__)
        template_file = os.path.join(path, "templates", "bids-app.json")

        with open(template_file) as f:
            template_string = f.read()

        errors = []
        app_name = os.path.basename(os.path.abspath(self.input_descriptor))
        version = "unknown"
        version_file = os.path.join(self.input_descriptor, "version")
        if os.path.exists(version_file):
            with open(version_file) as f:
                version = f.read().strip()
        git_repo = "https://github.com/BIDS-Apps/" + app_name
        entrypoint = self.get_entry_point(self.input_descriptor)
        container_image = "bids/" + app_name
        analysis_types = 'participant", "group", "session'

        if not entrypoint:
            errors.append("No entrypoint found in container.")

        if len(errors):
            raise_error(ValidationError, "Invalid descriptor:\n" + "\n".join(errors))

        template_string = template_string.replace("@@APP_NAME@@", app_name)
        template_string = template_string.replace("@@VERSION@@", version)
        template_string = template_string.replace("@@GIT_REPO_URL@@", git_repo)
        template_string = template_string.replace("@@DOCKER_ENTRYPOINT@@", entrypoint)
        template_string = template_string.replace(
            "@@CONTAINER_IMAGE@@", container_image
        )
        template_string = template_string.replace("@@ANALYSIS_TYPES@@", analysis_types)

        with open(self.output_descriptor, "w") as f:
            f.write(template_string)

    @importCatcher()
    def import_cwl(self):
        import oyaml as yaml

        # Read the CWL descriptor
        with open(self.input_descriptor) as f:
            cwl_desc = json.loads(json.dumps(yaml.load(f, Loader=yaml.FullLoader)))

        # validate yaml descriptor?

        bout_desc = {}
        # Command line
        if cwl_desc.get("baseCommand") is None:
            raise_error(
                ImportError,
                "Cannot find baseCommand attribute, "
                "perhaps you passed a workflow document, "
                "this is not supported",
            )
        if type(cwl_desc["baseCommand"]) is list:
            command_line = ""
            for i in cwl_desc["baseCommand"]:
                command_line += i + " "
        else:
            command_line = cwl_desc["baseCommand"]
        if cwl_desc.get("arguments"):
            for i in cwl_desc["arguments"]:
                if type(i) is dict:
                    raise_error(ImportError, "Dict arguments not supported.")
                if "$(runtime." in i:
                    raise_error(
                        ImportError,
                        "Runtime parameters " " are not supported:" " " + i,
                    )
                command_line += i + " "
        boutiques_inputs = []

        # Inputs
        def position(x):
            if (
                type(x) is dict
                and x.get("inputBinding")
                and x["inputBinding"].get("position")
            ):
                return x["inputBinding"]["position"]
            return 0

        sorted_inputs = sorted(
            cwl_desc["inputs"], key=lambda x: (position(cwl_desc["inputs"][x]))
        )
        # Mapping between CQL and Boutiques input types
        boutiques_types = {
            "string": "String",
            "File": "File",
            "File?": "File",
            "boolean": "Flag",
            "int": "Number",
            # float type?
        }
        for cwl_input in sorted_inputs:
            bout_input = {}
            # Easy stuff
            bout_input["id"] = cwl_input  # perhaps 'idify' that
            cwl_in_obj = cwl_desc["inputs"][cwl_input]
            if type(cwl_in_obj) is dict and cwl_in_obj.get("name"):
                bout_input["name"] = cwl_in_obj["name"]
            else:
                bout_input["name"] = cwl_input
            value_key = f"[{cwl_input.upper()}]"
            if type(cwl_in_obj) is dict and cwl_in_obj.get("inputBinding") is not None:
                command_line += " " + value_key
            bout_input["value-key"] = value_key

            # CWL type parsing
            if type(cwl_in_obj) is dict:
                cwl_type = cwl_in_obj["type"]
            else:
                cwl_type = "string"
            if type(cwl_type) is dict:  # It must be an array
                if cwl_type["type"] != "array":
                    raise_error(
                        ImportError,
                        "Only 1-level nested "
                        "types of type"
                        " 'array' are supported (CWL input: {})".format(cwl_input),
                    )
                if cwl_type.get("inputBinding") is not None:
                    raise_error(
                        ImportError,
                        "Input bindings of "
                        "array elements "
                        "are not supported (CWL input: {})".format(cwl_input),
                    )
                cwl_type = cwl_type["items"]
                bout_input["list"] = True
            if type(cwl_type) != str:
                raise_error(ImportError, f"Unknown type: {str(cwl_type)}")
            boutiques_type = boutiques_types[
                cwl_type.replace("[]", "").replace("?", "")
            ]
            bout_input["type"] = boutiques_type
            if cwl_type == "int":
                bout_input["integer"] = True
            if "?" in cwl_type or boutiques_type == "Flag":
                bout_input["optional"] = True

            # CWL input binding
            if type(cwl_in_obj) is dict:
                cwl_input_binding = cwl_in_obj["inputBinding"]
            else:
                cwl_input_binding = {}
            if cwl_input_binding.get("prefix"):
                bout_input["command-line-flag"] = cwl_input_binding["prefix"]
                if (
                    not (cwl_input_binding.get("separate") is None)
                    and cwl_input_binding["separate"] is False
                ):
                    bout_input["command-line-flag-separator"] = ""
            boutiques_inputs.append(bout_input)

            # array types
            if cwl_type.endswith("[]"):
                bout_input["list"] = True
                if cwl_input_binding.get("itemSeparator"):
                    if cwl_input_binding["itemSeparator"] != " ":
                        raise_error(
                            ImportError,
                            "Array separators will not be "
                            "supported until #76 is implemented",
                        )

        # Outputs

        def resolve_glob(glob, boutiques_inputs):
            if not glob.startswith("$"):
                return glob
            if not glob.startswith("$(inputs."):
                raise_error(ImportError, "Unsupported reference: " + glob)
            input_id = glob.replace("$(inputs.", "").replace(")", "")
            for i in boutiques_inputs:
                if i["id"] == input_id:
                    return i["value-key"]
            raise_error(ImportError, "Unresolved reference" " in glob: " + glob)

        boutiques_outputs = []
        sorted_outputs = sorted(
            cwl_desc["outputs"],
            key=(lambda x: cwl_desc["outputs"][x].get("outputBinding")),
        )
        for cwl_output in sorted_outputs:
            bout_output = {}
            bout_output["id"] = cwl_output  # perhaps 'idify' that
            if cwl_desc["outputs"][cwl_output].get("name"):
                bout_output["name"] = cwl_desc["outputs"][cwl_output]["name"]
            else:
                bout_output["name"] = cwl_output
            cwl_out_binding = cwl_desc["outputs"][cwl_output].get("outputBinding")
            if cwl_out_binding and cwl_out_binding.get("glob"):
                glob = cwl_out_binding["glob"]
                bout_output["path-template"] = resolve_glob(glob, boutiques_inputs)
                cwl_out_obj = cwl_desc["outputs"][cwl_output]
                if type(cwl_out_obj.get("type")) is dict:
                    if (
                        cwl_out_obj["type"].get("type")
                        and cwl_out_obj["type"]["type"] == "array"
                    ):
                        bout_output["list"] = True
                    else:
                        raise_error(
                            ImportError,
                            "Unsupported output type: " + cwl_output["type"],
                        )
                boutiques_outputs.append(bout_output)
        # Boutiques descriptors have to have at least 1 output file
        if len(boutiques_outputs) == 0 or cwl_desc.get("stdout"):
            stdout = cwl_desc.get("stdout") or "stdout.txt"
            command_line += " > " + stdout
            boutiques_outputs.append(
                {
                    "id": "stdout",
                    "name": "Standard output",
                    "path-template": "stdout.txt",
                }
            )

        # Mandatory boutiques fields
        bout_desc["command-line"] = command_line
        if cwl_desc.get("doc"):
            bout_desc["description"] = cwl_desc.get("doc").replace(os.linesep, "")
        else:
            bout_desc["description"] = "Tool imported from CWL."
        bout_desc["inputs"] = boutiques_inputs
        # This may not be a great idea but not sure if CWL tools have names
        bout_desc["name"] = op.splitext(op.basename(self.input_descriptor))[0]
        bout_desc["output-files"] = boutiques_outputs
        bout_desc["schema-version"] = "0.5"
        bout_desc["tool-version"] = "unknown"  # perhaphs there's one in cwl

        # Hints and requirements
        def parse_req(req, req_type, bout_desc):
            # We could support InitialWorkDirRequiment, through config files
            if req_type == "DockerRequirement":
                container_image = {}
                container_image["type"] = "docker"
                container_image["index"] = "index.docker.io"
                container_image["image"] = req["dockerPull"]
                bout_desc["container-image"] = container_image
                return
            if req_type == "EnvVarRequirement":
                bout_envars = []
                for env_var in req["envDef"]:
                    bout_env_var = {}
                    bout_env_var["name"] = env_var
                    bout_env_var["value"] = resolve_glob(
                        req["envDef"][env_var], boutiques_inputs
                    )
                    bout_envars.append(bout_env_var)
                    bout_desc["environment-variables"] = bout_envars
                return
            if req_type == "ResourceRequirement":
                suggested_resources = {}
                if req.get("ramMin"):
                    suggested_resources["ram"] = req["ramMin"]
                if req.get("coresMin"):
                    suggeseted_resources["cpu-cores"] = req["coresMin"]
                bout_desc["suggested-resources"] = suggested_resources
                return
            if req_type == "InitialWorkDirRequirement":
                listing = req.get("listing")
                for entry in listing:
                    file_name = entry.get("entryname")
                    assert file_name is not None
                    template = entry.get("entry")
                    for i in boutiques_inputs:
                        if i.get("value-key"):
                            template = template.replace(
                                "$(inputs." + i["id"] + ")", i.get("value-key")
                            )
                    template = template.split(os.linesep)
                    assert template is not None
                    name = op.splitext(file_name)[0]
                    boutiques_outputs.append(
                        {
                            "id": name,
                            "name": name,
                            "path-template": file_name,
                            "file-template": template,
                        }
                    )
                return
            raise_error(ImportError, "Unsupported requirement: " + str(req))

        for key in ["requirements", "hints"]:
            if cwl_desc.get(key):
                for i in cwl_desc[key]:
                    parse_req(cwl_desc[key][i], i, bout_desc)

        # enum types?

        # Write descriptor
        with open(self.output_descriptor, "w") as f:
            f.write(json.dumps(customSortDescriptorByKey(bout_desc), indent=4))
        validate_descriptor(loadJson(self.output_descriptor))

        if self.input_invocation is None:
            return

        # Convert invocation
        def get_input(descriptor_inputs, input_id):
            for inp in descriptor_inputs:
                if inp["id"] == input_id:
                    return inp
            return False

        boutiques_invocation = {}
        with open(self.input_invocation) as f:
            cwl_inputs = yaml.load(f, Loader=yaml.FullLoader)
        for input_name in cwl_inputs:
            if get_input(bout_desc["inputs"], input_name)["type"] != "File":
                input_value = cwl_inputs[input_name]
            else:
                input_value = cwl_inputs[input_name]["path"]
            boutiques_invocation[input_name] = input_value
        with open(self.output_invocation, "w") as f:
            f.write(
                json.dumps(
                    customSortInvocationByInput(
                        boutiques_invocation, json.dumps(bout_desc)
                    ),
                    indent=4,
                )
            )
        boutiques.invocation(self.output_descriptor, "-i", self.output_invocation)

    def import_config(self):
        def _getConfigFileString():
            with open(self.input_descriptor) as configFile:
                return configFile.read()

        def _getPropertiesFromValue(value):
            # Generate input properties from the input's value
            if isinstance(value, bool):
                return {
                    "type": "Flag",
                    "command-line-flag": "--TEMP",
                    "optional": True,
                    "default-value": value,
                }
            if isinstance(value, float):
                return {
                    "type": "Number",
                    "integer": False,
                    "default-value": value,
                }
            elif isinstance(value, int):
                return {
                    "type": "Number",
                    "integer": True,
                    "default-value": value,
                }
            elif isinstance(value, str):
                if re.match(r"\/.*\.[\w:]+", value) is not None:
                    return {"type": "File", "default-value": value}
                return {"type": "String", "default-value": value}
            elif isinstance(value, list):
                # Check all elements of list to extract list properties
                elementProperties, props = {}, []
                for element in value:
                    props.append(_getPropertiesFromValue(element))
                if all([p["type"] == "Flag" for p in props]):
                    elementProperties["type"] = "Flag"
                    elementProperties["command-line-flag"] = "--TEMP"
                    elementProperties["optional"] = True
                elif all([p["type"] == "Number" for p in props]):
                    elementProperties["type"] = "Number"
                    elementProperties["integer"] = all([p["integer"] for p in props])
                elif all([p["type"] == "File" for p in props]):
                    elementProperties["type"] = "File"
                else:
                    elementProperties["type"] = "String"
                elementProperties["list"] = True
                elementProperties["default-value"] = value
                return elementProperties

        def _createNameFromID(id):
            name = [c for c in id]
            for idx, c in enumerate(name):
                if not c.isalnum():
                    name[idx] = " "
                if c.isupper():
                    name[idx] = " " + c.lower()
            # Join char list into string and concatenate white spaces
            return " ".join("".join(name).split()).title()

        def _getOutputConfigFileTemplate(configFileFormat):
            return {
                "id": "config_file",
                "name": "Configuration file",
                "value-key": "[CONFIG_FILE]",
                "path-template": f"config.{configFileFormat}",
                "file-template": [],
            }

        def _getInputsFromConfigDict(input_config):
            # Generate inputs based on input config file (as dict)
            desc_inputs = []
            for id, value in input_config.items():
                newInput = {"id": id, "name": _createNameFromID(id)}
                newInput.update(_getPropertiesFromValue(value))
                newInput["value-key"] = "[{}]".format(
                    newInput["name"].replace(" ", "_").upper()
                )
                desc_inputs.append(newInput)
            return desc_inputs

        def import_json(descriptor):
            input_config = loadJson(self.input_descriptor)
            imported_inputs = _getInputsFromConfigDict(input_config)
            descriptor["inputs"].extend(imported_inputs)

            # file-template formatting depends on the config file's format
            output_config_file = _getOutputConfigFileTemplate(configFileFormat)
            output_config_file["file-template"].append("{")
            for inp in descriptor["inputs"]:
                input_entry = f"\"{inp['id']}\": \"{inp['value-key']}\""
                if inp != descriptor["inputs"][-1]:
                    input_entry += ","
                output_config_file["file-template"].append(input_entry)
            output_config_file["file-template"].append("}")

            descriptor["output-files"].append(output_config_file)
            return descriptor

        @importCatcher()
        def import_toml(descriptor):
            import toml

            tomlString = _getConfigFileString()
            input_config = toml.loads(tomlString, _dict=collections.OrderedDict)
            imported_inputs = _getInputsFromConfigDict(input_config)
            descriptor["inputs"].extend(imported_inputs)

            # file-template formatting depends on the config file's format
            output_config_file = _getOutputConfigFileTemplate(configFileFormat)
            for inp in descriptor["inputs"]:
                output_config_file["file-template"].append(
                    f"\"{inp['id']}\"=\"{inp['value-key']}\""
                )

            descriptor["output-files"].append(output_config_file)
            return descriptor

        @importCatcher()
        def import_yaml(descriptor):
            import oyaml as yaml

            yamlString = _getConfigFileString()
            input_config = yaml.load(yamlString, Loader=yaml.FullLoader)
            imported_inputs = _getInputsFromConfigDict(input_config)
            descriptor["inputs"].extend(imported_inputs)

            # file-template formatting depends on the config file's format
            output_config_file = _getOutputConfigFileTemplate(configFileFormat)
            for inp in descriptor["inputs"]:
                output_config_file["file-template"].append(
                    f"\"{inp['id']}\": {inp['value-key']}"
                )

            descriptor["output-files"].append(output_config_file)
            return descriptor

        descriptor = loadJson(self.output_descriptor)
        # Remove inputs and output-files from descriptor
        # Not emptying inputs also works, can be removed if needed
        descriptor["inputs"], descriptor["output-files"] = [], []
        configFileFormat = self.input_descriptor.split(".")[-1].lower()

        # Config K:V pair loading changes depending on config file type
        # Populating descriptor with inputs and configuration file
        if configFileFormat == "json":
            descriptor = import_json(descriptor)
        elif configFileFormat == "toml":
            descriptor = import_toml(descriptor)
        elif configFileFormat == "yml":
            descriptor = import_yaml(descriptor)

        descriptor["command-line"] = "tool [CONFIG_FILE]"

        with open(self.output_descriptor, "w+") as output:
            output.write(json.dumps(customSortDescriptorByKey(descriptor), indent=4))


class Docopt_Importer:
    @importCatcher()
    def __init__(self, docopt_str, base_descriptor):
        from docopt import (
            AnyOptions,
            DocoptLanguageError,
            Option,
            TokenStream,
            formal_usage,
            parse_argv,
            parse_defaults,
            parse_pattern,
        )

        with open(base_descriptor) as base_desc:
            self.descriptor = collections.OrderedDict(json.load(base_desc))

        del self.descriptor["groups"]
        del self.descriptor["inputs"]
        del self.descriptor["output-files"]
        self.docopt_str = docopt_str
        self.dependencies = collections.OrderedDict()
        self.all_desc_and_type = collections.OrderedDict()
        self.unique_ids = []

        try:
            # docopt code snippet to extract args tree (pattern)
            # should run if docopt script is valid
            options = parse_defaults(docopt_str)

            self.pattern = parse_pattern(
                formal_usage(self._parse_section("usage:", docopt_str)[0]),
                options,
            )

            argv = parse_argv(
                TokenStream(sys.argv[1:], DocoptLanguageError),
                list(options),
                False,
            )
            pattern_options = set(self.pattern.flat(Option))

            for options_shortcut in self.pattern.flat(AnyOptions):
                doc_options = parse_defaults(docopt_str)
                options_shortcut.children = list(set(doc_options) - pattern_options)
            matched, left, collected = self.pattern.fix().match(argv)
        except Exception:
            os.remove(base_descriptor)
            raise_error(ImportError, "Invalid docopt script")

        self.loadDocoptDescription()
        self.loadDescriptionAndType()
        self.generateInputsAndCommandLine(self.pattern)
        self.addInputsRecursive(self.dependencies)
        self.determineOptionality()
        self.createRootOneIsRequiredGroup()

    def loadDocoptDescription(self):
        # Get description of the script itself
        # which is everything except for usage and arguments
        self.descriptor["description"] = (
            self.docopt_str.replace(
                "".join(self._parse_section("usage:", self.docopt_str)), ""
            )
            .replace("".join(self._parse_section("arguments:", self.docopt_str)), "")
            .replace("".join(self._parse_section("options:", self.docopt_str)), "")
            .replace("\n\n", "\n")
            .strip()
        )

    @importCatcher()
    def loadDescriptionAndType(self):
        from docopt import Argument, Option

        # using docopt code to extract description and type from args
        for line in self._parse_section(
            "arguments:", self.docopt_str
        ) + self._parse_section("options:", self.docopt_str):
            _, _, s = line.partition(":")  # get rid of "options:"
            split = (
                re.split(r"\n[ \t]*(-\S+?)", "\n" + s)[1:]
                if line in self._parse_section("options:", self.docopt_str)
                else re.split(r"\n[ \t]*(<\S+?)", "\n" + s)[1:]
            )
            split = [s1 + s2 for s1, s2 in zip(split[::2], split[1::2])]
            # parse each line of Arguments and Options
            for arg_str in [
                s for s in split if (s.startswith("-") or s.startswith("<"))
            ]:
                arg = (
                    Option.parse(arg_str)
                    if arg_str.startswith("-")
                    else Argument.parse(arg_str)
                )
                arg_segs = arg_str.partition("  ")
                # Add desc and type to all_desc_and_type
                # key is initial id/name
                self.all_desc_and_type[arg.name] = {
                    "desc": arg_segs[-1].replace("\n", " ").replace("  ", "").strip()
                }
                if (
                    hasattr(arg, "value")
                    and arg.value is not None
                    and arg.value is not False
                ):
                    self.all_desc_and_type[arg.name]["default-value"] = arg.value
                if type(arg) is Option and arg.argcount > 0:
                    for typ in [
                        seg
                        for seg in arg_segs[0]
                        .replace(",", " ")
                        .replace("=", " ")
                        .split()
                        if seg[0] != "-"
                    ]:
                        self.all_desc_and_type[arg.name]["type"] = typ

    def generateInputsAndCommandLine(self, node):
        # Add commands and inputs, depth first
        child_node_type = type(node.children[0]).__name__
        if hasattr(node, "children") and (
            child_node_type == "Either" or child_node_type == "Required"
        ):
            for child in node.children:
                self.generateInputsAndCommandLine(child)
        else:
            # Traversing reached usage level, add script command
            self.descriptor["command-line"] = (
                self._parse_section("usage:", self.docopt_str)[0]
                .split("\n")[1:][0]
                .split()[0]
            )
            self._loadInputsFromUsage(node)

    def _loadInputsFromUsage(self, usage):
        from docopt import Argument

        ancestors = []
        for arg in usage.children:
            # Traverse usage args and add them to dependencies tree
            arg_type = type(arg).__name__
            if hasattr(arg, "children"):
                fchild_type = type(arg.children[0]).__name__
                # Has sub-arguments, maybe recurse into _loadRtrctnsFrmUsg
                # but have to deal with children in subtype
                if arg_type == "Optional" and fchild_type == "AnyOptions":
                    for option in arg.children[0].children:
                        self._addArgumentToDependencies(
                            option, ancestors=ancestors, optional=True
                        )
                elif arg_type == "OneOrMore":
                    list_name = f"<list_of_{self._getParamName(arg.children[0].name)}>"
                    list_arg = Argument(list_name)
                    list_arg.parse(list_name)
                    self.all_desc_and_type[list_name] = {
                        "desc": f"List of {self._getParamName(arg.children[0].name)}"
                    }
                    self._addArgumentToDependencies(
                        list_arg, ancestors=ancestors, isList=True
                    )
                    ancestors.append(list_name)
                elif arg_type == "Optional" and fchild_type == "Option":
                    for option in arg.children:
                        self._addArgumentToDependencies(
                            option, ancestors=ancestors, optional=True
                        )
                elif arg_type == "Optional" and fchild_type == "Either":
                    # Mutex choices group, add group to dependencies tree
                    # add choices args to group
                    ancestors = self._addGroupArgumentToDependencies(
                        arg, ancestors, optional=True
                    )
                elif arg_type == "Required" and fchild_type == "Either":
                    # Mutex choices group, add group to dependencies tree
                    # add choices args to group
                    ancestors = self._addGroupArgumentToDependencies(arg, ancestors)
            elif arg_type == "Command":
                self._addArgumentToDependencies(arg, ancestors=ancestors)
                ancestors.append(arg.name)
            elif arg_type == "Argument":
                self._addArgumentToDependencies(arg, ancestors=ancestors)
                ancestors.append(arg.name)
            elif arg_type == "Option":
                self._addArgumentToDependencies(arg, ancestors=ancestors, optional=True)
                ancestors.append(arg.name)
            else:
                raise_error(ImportError, f"Non implemented docopt arg.type: {arg_type}")

    def _addArgumentToDependencies(
        self, node, ancestors=None, isList=False, optional=False, members=[]
    ):
        # Add arg to dependency tree, establish hierarchy (parent/children)
        # by using ancestry of node
        p_node = self._getDependencyParentNode(ancestors)
        argAdded = self._generateDependencyArgument(node, isList, optional, members)

        if ancestors == [] and node.name not in self.dependencies:
            self.dependencies[node.name] = argAdded
        elif ancestors != [] and p_node is not None:
            argAdded["parent"] = p_node
            if node.name in p_node["children"]:
                p_node["children"][node.name]["children"].update(argAdded["children"])
            else:
                p_node["children"][node.name] = argAdded

        if hasattr(node, "long") and node.long is not None:
            # ensure flag has long hand flag
            argAdded["flag"] = node.long
            if (
                p_node is not None
                and "flag" in p_node
                and argAdded["flag"] == p_node["flag"]
            ):
                # if parent and child are same option (therefore has short-hand)
                # create new input with short-hand flag
                self.dependencies[p_node["children"][node.name]["name"]] = {
                    "id": p_node["children"][node.name]["id"],
                    "name": p_node["children"][node.name]["name"],
                    "desc": p_node["children"][node.name]["desc"],
                    "flag": node.short,
                    "optional": p_node["children"][node.name]["optional"],
                    "parent": None,
                    "children": p_node["children"][node.name]["children"],
                }
                del p_node["children"][node.name]

    def _generateDependencyArgument(
        self, node, isList=False, optional=False, members=[]
    ):
        # Generate & return dependency tree arg populated with parameters
        new_arg = {
            "id": node.name,
            "name": self._getUniqueId(self._getParamName(node.name)),
            "desc": (
                self.all_desc_and_type[node.name]["desc"]
                if node.name in self.all_desc_and_type
                else ""
            ),
            "optional": optional,
            "parent": None,
            "children": collections.OrderedDict(),
        }

        if new_arg is not None and isList:
            new_arg["isList"] = True
        if members != []:
            # dependency arg is group_arg if it contains mutex_members
            new_arg["mutex_members"] = members

        if node.name in self.all_desc_and_type:
            if "type" in self.all_desc_and_type[node.name]:
                new_arg["type"] = (
                    self._getStrippedName(self.all_desc_and_type[node.name]["type"])
                    if self.all_desc_and_type[node.name]["type"]
                    in {"File", "Flag", "Number", "String"}
                    or self.all_desc_and_type[node.name]["type"][1:-1]
                    in {"File", "Flag", "Number", "String"}
                    else "String"
                )

            # CODE OMITTED because of default-value bug in descriptor
            # if 'default-value' in self.all_desc_and_type[node.name]:
            # new_arg['default-value'] =\
            #    self.all_desc_and_type[node.name]['default-value']

        if hasattr(node, "long") and node.long is not None:
            # ensure flag has long hand flag
            new_arg["flag"] = node.long
        return new_arg

    @importCatcher()
    def _addGroupArgumentToDependencies(self, arg, ancestors, optional=False):
        from docopt import Argument

        # Add mutex choice group arg to dependency tree
        # group_arg contains members as dependency arguments
        options = arg.children[0].children
        p_node = self._getDependencyParentNode(ancestors)
        members = [self._generateDependencyArgument(option) for option in options]

        pretty_names = [member["name"] for member in members]
        names = [arg.name for arg in arg.children[0].children]
        gdesc = f"Group key for mutex choices: {' and '.join(names)}"
        for name in [name for name in names if name in self.all_desc_and_type]:
            gdesc = gdesc.replace(
                name,
                f"{name} ({self.all_desc_and_type[name]['desc']})",
            )
        gname = f"<{'_'.join(pretty_names)}>"
        grp_arg = Argument(gname)
        grp_arg.parse(gname)
        self.all_desc_and_type[gname] = {"desc": gdesc}

        self._addArgumentToDependencies(
            grp_arg, ancestors=ancestors, optional=optional, members=members
        )
        ancestors.append(grp_arg.name)
        return ancestors

    def addInputsRecursive(self, node, requires=[]):
        # Traverse dependencies and add inputs/groups
        # depending on children nb.
        args_ids = list(node.keys())
        if len(args_ids) == 1:
            self._addInputOrMutexGroupToDescriptor(node[args_ids[0]], requires)
            self.addInputsRecursive(node[args_ids[0]]["children"], requires)
        elif len(args_ids) > 1:
            mutex_names = []
            for name in args_ids:
                # requires-inputs list is closest ancestor with siblings
                # and single lineage descendants because this node has siblings
                self._addInputOrMutexGroupToDescriptor(
                    node[name], requires, addInputWithLineage=True
                )
                self.addInputsRecursive(node[name]["children"], [node[name]["name"]])
                if not node[name]["optional"]:
                    mutex_names.append(node[name]["name"])
            if len(mutex_names) > 1:
                mtxgrp_name = self._addMutexGroup(mutex_names)
                if node[args_ids[0]]["parent"]:
                    # add mutex group to parent node's requires-inputs
                    p_name = node[args_ids[0]]["parent"]["name"]
                    inputs = {inp["id"]: inp for inp in self.descriptor["inputs"]}
                    if "requires-inputs" in inputs[p_name]:
                        inputs[p_name]["requires-inputs"].append(mtxgrp_name)
                    else:
                        inputs[p_name]["requires-inputs"] = [mtxgrp_name]

    def _addInputOrMutexGroupToDescriptor(
        self, node, requires, addInputWithLineage=False
    ):
        if "mutex_members" in node:
            # Add mutex inputs and mutex group if arg is group_arg
            for mutex_member in node["mutex_members"]:
                self._addInput(mutex_member, requires)
            self._addMutexGroup([member["name"] for member in node["mutex_members"]])
        else:
            self._addInput(
                node,
                requires
                + (self._getLineageChildren(node, []) if addInputWithLineage else []),
                isList=node["isList"] if "isList" in node else False,
            )

    def _addInput(self, arg, requires, isList=False):
        # Add input to descriptor with all parameters
        if "inputs" not in self.descriptor:
            self.descriptor["inputs"] = []

        param_key = arg["id"]
        param_name = arg["name"]
        new_inp = {
            "id": param_name.replace("-", "_"),
            "name": param_name.replace("_", " ").replace("-", " "),
            "description": arg["desc"],
            "optional": True,
            "value-key": f"[{param_name}]".upper(),
        }
        self.descriptor["command-line"] += f" {new_inp['value-key']}"

        if requires != []:
            new_inp["requires-inputs"] = requires
        # Only add list param when isList
        if isList:
            new_inp["list"] = True
        if "default-value" in arg:
            new_inp["default-value"] = arg["default-value"]
        if "flag" in arg:
            if "type" in arg:
                new_inp["type"] = arg["type"]
            else:
                new_inp["type"] = "Flag"
            new_inp["command-line-flag"] = arg["flag"]
        else:
            new_inp["type"] = "String"
        if "mutex_members" in arg:
            mutex_names = []
            for choice_key in arg["children"]:
                choice = arg["children"][choice_key]["id"]
                if choice in arg["mutex_members"]:
                    mutex_names.append(choice_key)
            if len(mutex_names) > 1:
                self._addMutexGroup(mutex_names)
        else:
            self.descriptor["inputs"].append(new_inp)

    def _addMutexGroup(self, arg_names):
        # Add mutex group to descriptor
        pretty_name = "_".join([self._getParamName(name) for name in arg_names])
        unique_name = self._getUniqueId(pretty_name)
        new_group = {
            "id": pretty_name,
            "name": "Mutex group with members: {}".format(
                ", ".join([self._getStrippedName(name) for name in arg_names])
            ),
            "members": [self._getParamName(name) for name in arg_names],
            "mutually-exclusive": True,
        }
        if "groups" not in self.descriptor:
            self.descriptor["groups"] = []
        else:
            # Check/flatten nested groups in descriptor
            # If member of new mutex group (A) is itself a group (B),
            # get members of (B) and add them to the first group (A).
            # Needed to reduce validation complexity and potential
            # circular references.
            groups = {g["id"]: g for g in self.descriptor["groups"]}
            nested_grps = []
            for member in new_group["members"]:
                if member in [g["id"] for g in self.descriptor["groups"]]:
                    nested_grps.append(member)
                    new_group["members"].extend(groups[member]["members"])
            for nested_g in nested_grps:
                new_group["members"].remove(nested_g)
            new_group["name"] = "Mutex group with members: {}".format(
                ", ".join(new_group["members"])
            )
        self.descriptor["groups"].append(new_group)
        return new_group["id"]

    def determineOptionality(self):
        # Change optionality of inputs to false,
        # only applies to single root input and its single lineage descendants
        if len(self.dependencies) == 1:
            single_childs = [self.dependencies[next(iter(self.dependencies))]["name"]]
            single_childs += self._getLineageChildren(
                self.dependencies[next(iter(self.dependencies))]
            )
            for required_inp in [
                inp for inp in self.descriptor["inputs"] if inp["id"] in single_childs
            ]:
                required_inp["optional"] = False

    def createRootOneIsRequiredGroup(self):
        # Create OIR group with root inputs
        OIR_ids = []
        for root_inp in self.dependencies:
            OIR_ids.append(self.dependencies[root_inp]["name"].replace("-", "_"))
        pretty_name = "_".join(OIR_ids)
        unique_name = self._getUniqueId(pretty_name)
        new_group = {
            "id": pretty_name,
            "name": f"One is required group with members: {', '.join(OIR_ids)}",
            "members": OIR_ids,
            "one-is-required": True,
            "mutually-exclusive": True,
        }
        if len(OIR_ids) > 1:
            if "groups" not in self.descriptor:
                self.descriptor["groups"] = []
            self.descriptor["groups"].append(new_group)

    def _getLineageChildren(self, node, descendants=[]):
        # Recursively gets all the single lineage descendants names of
        # a dependency tree node
        child_keys = list(node["children"].keys())
        if len(child_keys) == 1 and not node["children"][child_keys[0]]["optional"]:
            descendants.append(node["children"][child_keys[0]]["name"])
            self._getLineageChildren(
                node["children"][child_keys[0]], descendants=descendants
            )
        return descendants

    def _getDependencyParentNode(self, ancestors):
        # Get parent dependency node, determined by ordered list of ancestors
        last_node = None
        for ancestor in ancestors:
            if last_node is None:
                for root in self.dependencies:
                    if self.dependencies[root]["id"] == ancestor:
                        last_node = self.dependencies[root]
            elif ancestor in [
                last_node["children"][key]["id"] for key in last_node["children"]
            ]:
                last_node = last_node["children"][
                    [
                        {"id": last_node["children"][n]["id"], "key": n}
                        for n in last_node["children"]
                    ][-1]["key"]
                ]
            else:
                return None
        return last_node

    def _getUniqueId(self, name):
        # Generate, store and return a unique_id for inputs/groups
        # will store generated id and original id for reference
        id_count = 1
        while name + ("_" + str(id_count) if id_count > 1 else "") in self.unique_ids:
            id_count += 1
        new_unique_id = name + ("_" + str(id_count) if id_count > 1 else "")
        self.unique_ids.append(new_unique_id)
        return new_unique_id

    def _getStrippedName(self, name):
        # only parses --arg and <arg>
        if name[0] == "-" and name[1] == "-":
            return name[2:]
        elif name[0] == "<" and name[-1] == ">":
            return name[1:-1]
        else:
            return name

    def _getParamName(self, param):
        # returns descriptor id compliant name
        chars = ["<", ">", "[", "]", "(", ")"]
        if param[0] == "-" and param[1] == "-":
            param = param[2:].replace("-", "_")
            for char in chars:
                param = param.replace(char, "")
        elif param[0] == "<" and param[-1] == ">":
            param = param[1:-1].replace("-", "_")
            for char in chars:
                param = param.replace(char, "")
        else:
            for char in chars:
                param = param.replace(char, "")
        return param

    def _parse_section(self, name, source):
        # old docopt code snippet
        # needed to extract section from docopt script
        pattern = re.compile(
            "^([^\n]*" + name + "[^\n]*\n?(?:[ \t].*?(?:\n|$))*)",
            re.IGNORECASE | re.MULTILINE,
        )
        return [s.strip() for s in pattern.findall(source)]



import glob
import os
import shutil
import sys
from unittest import TestCase

import pytest
from boutiques_mocks import example_boutiques_tool

from boutiques import __file__ as bfile


class BaseTest(TestCase):
    dir = "."
    tests_dir = os.path.join(os.path.dirname(bfile), "tests")
    test_temp = os.path.join(os.path.split(os.path.split(bfile)[0])[0], "test_temp")
    example1_descriptor = os.path.join(
        os.path.dirname(bfile),
        "schema",
        "examples",
        "example1",
        "example1_docker.json",
    )

    @pytest.fixture(autouse=True)
    def clean_test_temp(self):
        # Clean test temp directory after every test
        yield
        if os.path.exists(self.test_temp):
            shutil.rmtree(self.test_temp)

        os.makedirs(self.test_temp, exist_ok=True)

    @pytest.fixture(autouse=True)
    def reset_mock_zenodo_record(self):
        example_boutiques_tool.reset()

    def setup(self, dir):
        self.dir = dir

    def get_file_path(self, file):
        return os.path.join(os.path.join(self.tests_dir, self.dir), file)

    def assert_nothing(self, ret):
        pass

    def assert_reflected_output(self, ret):
        self.assertIn("This is stdout", ret.stdout)
        self.assertIn("This is stderr", ret.stderr)

    def assert_reflected_output_nonutf8(self, ret):
        self.assertIn("\\xc7a c'est stdout", ret.stdout)
        self.assertIn("This is std\\xe9rr", ret.stderr)

    def assert_no_output(self, ret):
        self.assertEqual(ret.stdout, "")
        self.assertEqual(ret.stderr, "")

    def assert_only_stdout(self, ret):
        self.assertNotEqual(ret.stdout, "")
        self.assertEqual(ret.stderr, "")

    def assert_successful_return(
        self,
        ret,
        required_files=None,
        required_files_len=0,
        aditional_assertions=assert_nothing,
    ):

        aditional_assertions(ret)

        self.assertEqual(ret.error_message, "")
        self.assertEqual(ret.exit_code, 0)
        self.assertEqual(ret.missing_files, [])
        if required_files is not None:
            if len(required_files) > required_files_len:
                required_files_len = len(required_files)
            self.assertEqual(len(ret.output_files), required_files_len)
            for required_file in required_files:
                self.assertIn(required_file, [f.file_name for f in ret.output_files])

    def assert_failed_return(
        self,
        ret,
        exit_code,
        error_message,
        missing_files=[],
        missing_file_len=0,
    ):

        if len(missing_files) > missing_file_len:
            missing_file_len = len(missing_files)

        self.assertEqual(ret.error_message, error_message)
        self.assertEqual(ret.exit_code, exit_code)
        self.assertEqual(len(ret.missing_files), missing_file_len)
        for missing in missing_files:
            self.assertIn(missing, [f.file_name for f in ret.missing_files])



#!/usr/bin/env python

import os
import subprocess
from pathlib import Path
from shutil import copy2
from unittest import mock

import pytest
import simplejson as json
from boutiques_mocks import example_boutiques_tool, mock_get

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.descriptor2func import function
from boutiques.localExec import ExecutorError
from boutiques.tests.BaseTest import BaseTest
from boutiques.util.utils import LoadError


class TestExample1(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("example1")

    @pytest.fixture(autouse=True)
    def clean_up(self):
        fls = os.listdir("./")
        for fl in fls:
            if (fl.startswith("log") or fl.startswith("config")) and fl.endswith(
                ".txt"
            ):
                os.remove(fl)

    # Captures the stdout and stderr during test execution
    # and returns them as a tuple in readouterr()
    @pytest.fixture(autouse=True)
    def capture_st(self, capfd):
        self.capfd = capfd

    def test_example1_validate(self):
        self.assertIsNone(bosh.validate(self.example1_descriptor))

    def test_example1_no_exec(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        self.assert_successful_return(
            bosh.execute("simulate", self.example1_descriptor, "-i", invoc),
            aditional_assertions=self.assert_only_stdout,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        ret = bosh.execute(
            "launch",
            self.example1_descriptor,
            invoc,
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )

        # Make sure stdout and stderr are not printed on the fly
        # for non-streaming mode
        out, err = self.capfd.readouterr()
        self.assertNotIn("This is stdout", out)
        self.assertNotIn("This is stderr", err)

        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_debug(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.example1_descriptor,
                "--debug",
                invoc,
                "--skip-data-collection",
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_stream_output(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        ret = bosh.execute(
            "launch",
            self.example1_descriptor,
            "-s",
            invoc,
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )

        # Make sure stdout and stderr are printed on the fly for
        # streaming mode
        out, _ = self.capfd.readouterr()
        self.assertIn("This is stdout", out)
        self.assertIn("This is stderr", out)

        self.assertIsNone(ret.stdout)
        self.assertIsNone(ret.stderr)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_inv_as_json_obj(self):
        invocationStr = open(
            os.path.join(
                os.path.dirname(bfile),
                "schema",
                "examples",
                "example1",
                "invocation.json",
            )
        ).read()
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.example1_descriptor,
                invocationStr,
                "--skip-data-collection",
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_inv_as_json_obj_debug(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.example1_descriptor,
                "--debug",
                invoc,
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
                "--skip-data-collection",
            ),
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_desc_as_json_obj(self):
        descStr = open(self.example1_descriptor).read()
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        self.assert_successful_return(
            bosh.execute(
                "launch",
                descStr,
                invoc,
                "--skip-data-collection",
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_desc_as_json_obj_debug(self):
        descStr = open(self.example1_descriptor).read()
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        self.assert_successful_return(
            bosh.execute(
                "launch",
                descStr,
                invoc,
                "--debug",
                "--skip-data-collection",
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_json_string_invalid(self):
        invocationStr = open(self.get_file_path("invocation_invalid.json")).read()
        with pytest.raises(LoadError) as e:
            bosh.execute(
                "launch",
                self.example1_descriptor,
                "-u",
                invocationStr,
                "--skip-data-collection",
            )
        self.assertIn("Cannot parse input", str(e.getrepr(style="long")))

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch("requests.get", return_value=mock_get())
    def test_example1_exec_docker_from_zenodo(self, _):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        ret = bosh.execute(
            "launch",
            "zenodo." + str(example_boutiques_tool.id),
            invoc,
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )

        # Make sure stdout and stderr are not printed on the fly
        # for non-streaming mode
        out, err = self.capfd.readouterr()
        self.assertNotIn("This is stdout", out)
        self.assertNotIn("This is stderr", err)
        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch("requests.get", return_value=mock_get())
    def test_example1_exec_docker_from_zenodo_debug(self, _):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        self.assert_successful_return(
            bosh.execute(
                "launch",
                "zenodo." + str(example_boutiques_tool.id),
                "--debug",
                invoc,
                "--skip-data-collection",
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch("requests.get", return_value=mock_get())
    def test_example1_exec_docker_from_zenodo_desc2func_default(self, _):
        # No mode provided, defaults to 'launch'
        example_tool = function("zenodo." + str(example_boutiques_tool.id))
        ret = example_tool(
            str_input_list=["a", "b", "c"],
            str_input="coin;plop",
            file_input="./pyproject.toml",
            file_list_input=["./pyproject.toml"],
            list_int_input=[1, 2, 3],
            config_num=4,
            enum_input="val1",
        )
        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch("requests.get", return_value=mock_get())
    def test_example1_exec_docker_from_zenodo_desc2func_launch(self, _):
        # Launch mode
        example_tool = function("zenodo." + str(example_boutiques_tool.id))
        ret = example_tool(
            "launch",
            str_input_list=["a", "b", "c"],
            str_input="coin;plop",
            file_input="./pyproject.toml",
            file_list_input=["./pyproject.toml"],
            list_int_input=[1, 2, 3],
            config_num=4,
            enum_input="val1",
        )
        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch("requests.get", return_value=mock_get())
    def test_example1_exec_docker_from_zenodo_desc2func_simulate(self, _):
        # Simulate with invocation
        example_tool = function("zenodo." + str(example_boutiques_tool.id))
        ret = example_tool(
            "simulate",
            str_input_list=["a", "b", "c"],
            str_input="coin;plop",
            file_input="./pyproject.toml",
            file_list_input=["./pyproject.toml"],
            list_int_input=[1, 2, 3],
            config_num=4,
            enum_input="val1",
        )
        self.assert_successful_return(ret, aditional_assertions=self.assert_only_stdout)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch("requests.get", return_value=mock_get())
    def test_example1_exec_docker_from_zenodo_desc2func_simNoInvoc(self, _):
        # Simulate without invocation
        example_tool = function("zenodo." + str(example_boutiques_tool.id))
        ret = example_tool("simulate")
        self.assertIn("exampleTool1.py -c", ret.stdout)

    @pytest.mark.xfail(reason="Travis to GH action transition")
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_example1_exec_singularity(self):
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.get_file_path("example1_sing.json"),
                self.get_file_path("invocation_sing.json"),
                "--skip-data-collection",
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.xfail(reason="Travis to GH action transition")
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_example1_exec_singularity_debug(self):
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.get_file_path("example1_sing.json"),
                "--debug",
                self.get_file_path("invocation_sing.json"),
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4.txt"],
            2,
            self.assert_reflected_output,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_example1_crash_pull_singularity(self):
        with pytest.raises(ExecutorError) as e:
            bosh.execute(
                "launch",
                self.get_file_path("example1_sing_crash_pull.json"),
                self.get_file_path("invocation_sing.json"),
                "--skip-data-collection",
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            )
        self.assertIn("Could not pull Singularity image", str(e.getrepr(style="long")))

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_missing_script(self):
        be = bosh.execute(
            "launch",
            self.example1_descriptor,
            self.get_file_path("invocation_missing_script.json"),
            "--skip-data-collection",
            "--no-automounts",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )
        self.assert_failed_return(
            be, 2, "File does not exist!", ["./test_temp/log-4-pwet.txt"], 1
        )

    @pytest.mark.skip(reason="Fails after transition to GH action")
    def test_example1_exec_fail_cli(self):
        command = (
            "bosh",
            "exec",
            "launch",
            self.example1_descriptor,
            self.get_file_path("invocation_missing_script.json"),
            "--skip-data-collection",
            "--no-automounts",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )
        command = " ".join(command)
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        process.communicate()
        assert process.returncode == 2, command

    def test_example1_no_exec_random(self):
        self.assert_successful_return(
            bosh.execute("simulate", self.example1_descriptor),
            aditional_assertions=self.assert_only_stdout,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_non_utf8(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        ret = bosh.execute(
            "launch",
            self.get_file_path("example1_docker_nonutf8.json"),
            invoc,
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )

        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output_nonutf8,
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_docker_non_utf8_debug(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        self.assert_successful_return(
            bosh.execute(
                "launch",
                self.get_file_path("example1_docker_nonutf8.json"),
                "--debug",
                invoc,
                "-v",
                f"{self.get_file_path('example1_mount1')}:/test_mount1",
                "-v",
                f"{self.get_file_path('example1_mount2')}:/test_mount2",
            ),
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output_nonutf8,
        )

    @pytest.mark.xfail(reason="Travis to GH action transition")
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_example1_exec_docker_force_singularity(self):
        ret = bosh.execute(
            "launch",
            self.example1_descriptor,
            self.get_file_path("invocation_no_opts.json"),
            "--skip-data-collection",
            "--force-singularity",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )

        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )
        self.assertIn("Local (boutiques-example1-test.simg)", ret.container_location)
        self.assertIn("singularity exec", ret.container_command)

    @pytest.mark.xfail(reason="Travis to GH action transition")
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_example1_exec_docker_Index_force_singularity(self):
        ret = bosh.execute(
            "launch",
            self.get_file_path("example1_docker_w_index.json"),
            self.get_file_path("invocation_no_opts.json"),
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
            "--force-singularity",
        )

        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )
        self.assertIn("Local (boutiques-example1-test.simg)", ret.container_location)
        self.assertIn("singularity exec", ret.container_command)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_exec_singularity_force_docker(self):
        ret = bosh.execute(
            "launch",
            self.get_file_path("example1_sing.json"),
            self.get_file_path("invocation_sing_no_opts.json"),
            "--skip-data-collection",
            "--force-docker",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )

        self.assert_successful_return(
            ret, ["./test_temp/log-4.txt"], 2, self.assert_reflected_output
        )
        self.assertIn("Pulled from Docker", ret.container_location)
        self.assertIn("docker run", ret.container_command)

    def docker_not_installed(command):
        if command == "docker":
            return False
        else:
            return True

    @pytest.mark.xfail(reason="Travis to GH action transition")
    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    @mock.patch(
        "boutiques.localExec.LocalExecutor._isCommandInstalled",
        side_effect=docker_not_installed,
    )
    def test_example1_exec_docker_not_installed(self, mock_docker_not_installed):
        ret = bosh.execute(
            "launch",
            self.example1_descriptor,
            self.get_file_path("invocation_no_opts.json"),
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )
        self.assert_successful_return(
            ret,
            ["./test_temp/log-4-coin;plop.txt"],
            2,
            self.assert_reflected_output,
        )
        self.assertIn("Local (boutiques-example1-test.simg)", ret.container_location)
        self.assertIn("singularity exec", ret.container_command)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_conditional_outputFiles_created(self):
        os.makedirs(self.test_temp, exist_ok=True)
        ex = bosh.execute(
            "launch",
            self.get_file_path("example1_docker_conditional_outputFiles.json"),
            self.get_file_path("example1_conditional_invoc.json"),
            "--skip-data-collection",
        )

        outFileList = [str(out) for out in ex.output_files]
        self.assertIn(
            "./test_temp/TEST.one.three.two_out1.txt" " (out1, Required)",
            outFileList,
        )
        self.assertIn("./test_temp/TEST_string1.txt (out2, Required)", outFileList)
        self.assertEqual([], ex.missing_files)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_environment_variables_from_invoc(self):
        Path(self.test_temp, "test_path.d").touch()
        ex = bosh.execute(
            "launch",
            self.get_file_path("example1_envVars_from_inputs.json"),
            self.get_file_path("test_input_env_var_invoc.json"),
            "--skip-data-collection",
        )

        self.assertIn("./test_temp/test_path.d", ex.stdout)

    @pytest.mark.usefixtures("skip_if_no_docker")
    def test_missing_mount_location(self):
        from boutiques.localExec import ExecutorError

        with pytest.raises(ExecutorError):
            bosh.execute(
                "launch",
                self.get_file_path("example1_envVars_from_inputs.json"),
                self.get_file_path("test_input_env_var_invoc.json"),
                "--skip-data-collection",
            )

    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_example1_exec_container_image_contains_index(self):
        ret = bosh.execute(
            "launch",
            self.get_file_path("conImage_with_index.json"),
            self.get_file_path("input_invoc.json"),
            "--skip-data-collection",
        )

        self.assertIn("Local (boutiques-example1-test.simg)", ret.container_location)
        self.assertIn("singularity exec", ret.container_command)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_autoMount_input_files(self):
        base_path = self.get_file_path("automount")
        # Test files must be created outside of [...]/tools/python/
        # because it is mounted by default
        test_dir = os.path.split(os.path.split(bfile)[0])[0]
        copy2(os.path.join(base_path, "file1.txt"), test_dir)
        copy2(os.path.join(base_path, "file2.txt"), test_dir)
        copy2(os.path.join(base_path, "file3.txt"), test_dir)
        # copy2(os.path.join(test_invoc), test_dir)
        invoc_dict = {
            "file": "./file1.txt",
            "file_list": ["./file2.txt", "./file3.txt"],
        }
        # Create test invoc based on absolute test_dir path
        test_invoc = self.get_file_path("test_automount_invoc.json")
        with open(test_invoc, "w+") as invoc:
            invoc.write(json.dumps(invoc_dict))

        ex = bosh.execute(
            "launch",
            self.get_file_path("test_automount_desc.json"),
            test_invoc,
            "--skip-data-collection",
        )

        try:
            self.assertIn("Hello, World!", ex.stdout.replace("\n", ""))
        finally:
            os.remove(os.path.join(test_dir, "file1.txt"))
            os.remove(os.path.join(test_dir, "file2.txt"))
            os.remove(os.path.join(test_dir, "file3.txt"))
            os.remove(test_invoc)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_autoMount_none(self):
        base_path = self.get_file_path("automount")
        test_invoc = self.get_file_path("test_automount_invoc.json")
        # Test files must be created outside of [...]/tools/python/
        # because it is mounted by default
        test_dir = os.path.split(os.path.split(bfile)[0])[0]
        copy2(os.path.join(base_path, "file1.txt"), test_dir)
        copy2(os.path.join(base_path, "file2.txt"), test_dir)
        copy2(os.path.join(base_path, "file3.txt"), test_dir)
        invoc_dict = {
            "file": test_dir + "./file1.txt",
            "file_list": [test_dir + "./file2.txt", test_dir + "./file3.txt"],
        }
        # Create test invoc based on absolute test_dir path
        with open(test_invoc, "w+") as invoc:
            invoc.write(json.dumps(invoc_dict))

        ex = bosh.execute(
            "launch",
            self.get_file_path("test_automount_desc.json"),
            test_invoc,
            "--no-automounts",
            "--skip-data-collection",
        )

        try:
            self.assertIn("file1.txt: No such file or directory", ex.stderr)
            self.assertIn("file2.txt: No such file or directory", ex.stderr)
            self.assertIn("file3.txt: No such file or directory", ex.stderr)
        finally:
            os.remove(os.path.join(test_dir, "file1.txt"))
            os.remove(os.path.join(test_dir, "file2.txt"))
            os.remove(os.path.join(test_dir, "file3.txt"))
            os.remove(test_invoc)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_example1_container_opts(self):
        container_opts = "-e TEST_CONTAINER_OPTS=TEST_CONTAINER_OPTS"
        extra_container_opts = "-p 8888:8888 -e MORE_ENV=MORE_ENV"

        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        ret = bosh.execute(
            "launch",
            self.example1_descriptor,
            invoc,
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
            "--container-opts",
            container_opts,
            "--container-opts",
            extra_container_opts,
        )

        self.assertIn(container_opts, ret.container_command)
        self.assertIn(extra_container_opts, ret.container_command)

    @pytest.mark.skipif(
        subprocess.Popen("type singularity", shell=True).wait(),
        reason="Singularity not installed",
    )
    def test_example1_singularity_container_opts(self):
        # check that container runtime options are not ignored when specified on
        # command-line, and the container runtime differs from the descriptor
        container_opts = "--cpus 1"

        ret = bosh.execute(
            "launch",
            self.example1_descriptor,
            self.get_file_path("invocation_no_opts.json"),
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
            "--force-singularity",
            "--container-opts",
            container_opts,
        )

        self.assertIn(container_opts, ret.container_command)



#!/usr/bin/env python

import os

import simplejson as json

from boutiques.logger import print_info, raise_error
from boutiques.util.utils import customSortDescriptorByKey, loadJson
from boutiques.validator import ValidationError, validate_descriptor
from boutiques.zenodoHelper import ZenodoError, ZenodoHelper


class Publisher:

    def __init__(
        self,
        descriptor_file_name,
        auth_token,
        verbose=False,
        sandbox=False,
        no_int=False,
        replace=False,
        id=None,
    ):
        # Straightforward assignments
        self.verbose = verbose
        self.sandbox = sandbox
        self.descriptor_file_name = descriptor_file_name
        self.no_int = no_int
        self.zenodo_access_token = auth_token
        self.zenodo_helper = ZenodoHelper(sandbox, no_int, verbose)
        self.descriptor = loadJson(self.descriptor_file_name)

        # remove zenodo prefix of ID to update
        try:
            self.id_to_update = id.split(".", 1)[1] if id else None
        except IndexError:
            raise_error(
                ZenodoError,
                "Zenodo ID must be prefixed by " "'zenodo', e.g. zenodo.123456",
            )

        # Validate and load descriptor
        validate_descriptor(self.descriptor)

        # Get relevant descriptor properties
        self.url = self.descriptor.get("url")
        self.tool_doi = self.descriptor.get("tool-doi")
        self.descriptor_url = self.descriptor.get("descriptor-url")
        self.online_platforms = self.descriptor.get("online-platform-urls")

        # Get tool author and check that it's defined
        if self.descriptor.get("author") is None:
            raise_error(
                ZenodoError,
                "Tool must have an author to be "
                "published. Add an 'author' property to your "
                "descriptor.",
            )
        self.creator = self.descriptor["author"]

        # Get tool container and check that it's defined
        if self.descriptor.get("container-image") is None:
            raise_error(
                ZenodoError,
                "Tool must have a container image to be "
                "published. Add a 'container-image' property to your "
                "descriptor.",
            )

        # If in replace mode, make sure descriptor has a DOI and get the ID.
        # Otherwise, make sure the descriptor does not have a DOI.
        if replace:
            if self.descriptor.get("doi") is None:
                raise_error(
                    ZenodoError,
                    "To publish an updated version of a "
                    "previously published descriptor, the descriptor "
                    "must contain a DOI. This DOI will be replaced "
                    "with a new one.",
                )
            else:
                self.id_to_update = self.descriptor.get("doi").split(".")[-1]
        elif self.descriptor.get("doi") is not None:
            raise_error(
                ZenodoError,
                "Descriptor already has a DOI. Please "
                "remove it from the descriptor before publishing it "
                "again, or use the --replace flag to publish an "
                "updated version. A new DOI will be generated.",
            )

        self.config_file = os.path.join(os.path.expanduser("~"), ".boutiques")

        # Fix Zenodo access token
        self.zenodo_access_token = self.zenodo_helper.verify_zenodo_access_token(
            self.zenodo_access_token
        )

        # Set Zenodo endpoint
        self.zenodo_endpoint = self.zenodo_helper.get_zenodo_endpoint()

    def publish(self):
        if not self.no_int:
            prompt = (
                "The descriptor will be published to Zenodo, "
                "this cannot be undone. Are you sure? (Y/n) "
            )
            ret = input(prompt)
            if ret.upper() != "Y":
                return

        if self.id_to_update is not None:
            publish_update = True
        else:
            # perform a search to check if descriptor is an updated version
            # of an existing one
            from boutiques.searcher import Searcher

            searcher = Searcher(
                self.descriptor.get("name"),
                self.verbose,
                self.sandbox,
                exact_match=True,
            )
            r = self.zenodo_helper.zenodo_search(searcher.query, searcher.query_line)

            publish_update = False
            for hit in r.json()["hits"]["hits"]:
                title = hit["metadata"]["title"]
                if title == self.descriptor.get("name"):
                    self.id_to_update = hit["id"]
                    break

            if self.id_to_update is not None:
                if not self.no_int:
                    prompt = (
                        "Found an existing record with the same name, "
                        "would you like to update it? "
                        "(Y:Update existing / n:Publish new entry with "
                        "name {}) ".format(self.descriptor.get("name"))
                    )
                    ret = input(prompt)
                    if ret.upper() == "Y":
                        publish_update = True
                else:
                    publish_update = True

        if publish_update:
            deposition_id = self.zenodo_helper.zenodo_deposit_updated_version(
                self.create_metadata(),
                self.zenodo_access_token,
                self.id_to_update,
            )
        else:
            deposition_id = self.zenodo_helper.zenodo_deposit(
                self.create_metadata(), self.zenodo_access_token
            )

        # If in replace mode, remove the old DOI
        if self.descriptor.get("doi"):
            del self.descriptor["doi"]
        with open(self.descriptor_file_name, "w") as fhandle:
            fhandle.write(json.dumps(self.descriptor, indent=4))

        self.zenodo_helper.zenodo_upload_file(
            deposition_id,
            self.descriptor_file_name,
            zenodo_access_token=self.zenodo_access_token,
            error_msg="Cannot upload descriptor to Zenodo",
            verbose_msg="Descriptor uploaded to Zenodo",
        )
        self.doi = self.zenodo_helper.zenodo_publish(
            self.zenodo_access_token, deposition_id, "Descriptor"
        )

        # Assign new doi to published descriptor
        self.descriptor["doi"] = self.doi
        with open(self.descriptor_file_name, "w") as fhandle:
            fhandle.write(json.dumps(self.descriptor, indent=4))

        if os.path.isfile(self.descriptor_file_name):
            return "OK"
        return False

    def create_metadata(self):
        data = {
            "metadata": {
                "title": self.descriptor["name"],
                "upload_type": "software",
                "description": self.descriptor["description"]
                or f"Boutiques descriptor for {self.descriptor['name']}",
                "creators": [{"name": self.creator}],
                "version": self.descriptor["tool-version"],
                "keywords": [
                    "Boutiques",
                    f"schema-version:{self.descriptor['schema-version']}",
                ],
            }
        }
        keywords = data["metadata"]["keywords"]
        if self.descriptor.get("tags"):
            for key, value in self.descriptor.get("tags").items():
                # Tag is of form 'tag-name': true, it is a single-string
                if isinstance(value, bool):
                    keywords.append(key)
                # Tag is of form 'tag-name':'tag-value', it is a key-value pair
                if isinstance(value, str):
                    keywords.append(key + ":" + value)
                # Tag is of form 'tag-name': ['value1', 'value2'], it is a
                # list of key-value pairs
                elif isinstance(value, list):
                    keywords += [key + ":" + item for item in value]
        if self.descriptor.get("container-image"):
            keywords.append(self.descriptor["container-image"]["type"])
        if self.descriptor.get("tests"):
            keywords.append("tested")
        if self.descriptor.get("deprecated-by-doi"):
            keywords.append("deprecated")
            if isinstance(self.descriptor["deprecated-by-doi"], str):
                keywords.append(
                    "deprecated-by-doi:" + self.descriptor["deprecated-by-doi"]
                )
            self.addRelatedIdentifiers(
                data,
                self.descriptor["deprecated-by-doi"],
                "isPreviousVersionOf",
            )

        if self.url is not None:
            self.addRelatedIdentifiers(data, self.url, "hasPart")
        if self.online_platforms is not None:
            for p in self.online_platforms:
                self.addRelatedIdentifiers(data, p, "hasPart")
        if self.tool_doi is not None:
            self.addRelatedIdentifiers(data, self.tool_doi, "hasPart")
        if self.descriptor_url is not None:
            self.addRelatedIdentifiers(data, self.descriptor_url, "hasPart")
        return data

    def addRelatedIdentifiers(self, data, identifier, relation):
        if data["metadata"].get("related_identifiers") is None:
            data["metadata"]["related_identifiers"] = []
        data["metadata"]["related_identifiers"].append(
            {"identifier": identifier, "relation": relation}
        )



import json
import os
from unittest import mock

from boutiques import __file__ as bfile
from boutiques.bosh import bosh
from boutiques.deprecate import DeprecateError, deprecate
from boutiques.tests.BaseTest import BaseTest
from boutiques.tests.boutiques_mocks import (
    MockHttpResponse,
    example_boutiques_tool,
    get_zenodo_record,
    mock_zenodo_delete_files,
    mock_zenodo_search,
    mock_zenodo_test_api,
)
from boutiques.util.utils import loadJson

ZENODO_SANDBOX_TOKEN = "fake-token-123"


def mock_get(*args, **kwargs):
    # Check that URL looks good
    split = args[0].split("/")
    assert len(split) >= 5

    command = split[4]
    # Records command
    if command == "records":
        assert len(split) >= 6
        record_id = split[12] if len(split) > 6 else split[5]
        if record_id == "00000":
            # Inexistent tool
            return MockHttpResponse(404)
        mock_record1 = example_boutiques_tool
        mock_record1.id = int(record_id)
        if record_id == "22222":
            mock_record1.is_last_version = False
        mock_get = get_zenodo_record(mock_record1)
        if "?q=keywords:" in args[0]:
            return mock_zenodo_search([mock_record1])
        return MockHttpResponse(200, mock_get)

    # Deposit command
    if command == "deposit":
        # Check auth
        if kwargs.get("params") and kwargs.get("params").get("access_token"):
            return MockHttpResponse(200)
        else:
            return MockHttpResponse(403)


def mock_post(*args, **kwargs):
    # Check that URL looks good
    split = args[0].split("/")
    assert len(split) >= 5
    command = split[4]
    # Deposit command
    if command == "deposit":
        json = {
            "links": {"latest_draft": "plop/coin/pan/12345"},
            "doi": "foo/bar",
            "files": [{"id": "qwerty"}],
        }
        if args[0].endswith("actions/publish"):
            return MockHttpResponse(202, json)
        else:
            return MockHttpResponse(201, json)


def mock_download_deprecated(url, file_path):
    # Mocks the download and save of a deprecated descriptor
    example_1_path = os.path.join(
        os.path.dirname(bfile),
        "schema",
        "examples",
        "example1",
        "example1_docker.json",
    )
    example_1_json = loadJson(example_1_path)
    example_1_json["deprecated-by-doi"] = "a_doi"
    cache_dir = os.path.join(
        os.path.expanduser("~"), ".cache", "boutiques", "production"
    )
    with open(file_path, "w") as f:
        f.write(json.dumps(example_1_json))
    return [f.name]


@mock.patch("requests.get", side_effect=mock_get)
@mock.patch("requests.post", side_effect=mock_post)
@mock.patch("requests.put", side_effect=mock_zenodo_test_api)
@mock.patch("requests.delete", side_effect=mock_zenodo_delete_files)
class TestDeprecate(BaseTest):
    def test_deprecate(self, *args):
        new_doi = bosh(
            [
                "deprecate",
                "--verbose",
                "--by",
                "zenodo.12345",
                "zenodo." + str(example_boutiques_tool.id),
                "--zenodo-token",
                ZENODO_SANDBOX_TOKEN,
            ]
        )
        self.assertTrue(new_doi)

    def test_deprecate_no_by(self, *args):
        new_doi = bosh(
            [
                "deprecate",
                "--verbose",
                "zenodo." + str(example_boutiques_tool.id),
                "--zenodo-token",
                ZENODO_SANDBOX_TOKEN,
            ]
        )
        self.assertTrue(new_doi)

    def test_deprecate_by_inexistent(self, *args):
        with self.assertRaises(DeprecateError) as e:
            new_doi = bosh(
                [
                    "deprecate",
                    "--verbose",
                    "--by",
                    "zenodo.00000",
                    "zenodo." + str(example_boutiques_tool.id),
                    "--zenodo-token",
                    ZENODO_SANDBOX_TOKEN,
                ]
            )
        self.assertTrue("Tool does not exist" in str(e.exception))

    def test_deprecate_deprecated(self, *args):
        new_doi = deprecate(
            zenodo_id="zenodo.11111",
            sandbox=True,
            verbose=True,
            zenodo_token=ZENODO_SANDBOX_TOKEN,
            download_function=mock_download_deprecated,
        )
        self.assertTrue(new_doi)

    def test_deprecate_previous_version(self, *args):
        with self.assertRaises(DeprecateError) as e:
            new_doi = bosh(
                [
                    "deprecate",
                    "--verbose",
                    "zenodo.22222",
                    "--zenodo-token",
                    ZENODO_SANDBOX_TOKEN,
                ]
            )
        self.assertTrue("Tool zenodo.22222 has a newer version" in str(e.exception))



#!/usr/bin/env python

import argparse
import os
import os.path as op
import sys

import simplejson as json
from jsonschema import ValidationError
from tabulate import tabulate

from boutiques.boshParsers import *
from boutiques.dataHandler import DataHandlerError
from boutiques.exporter import ExportError
from boutiques.importer import ImportError
from boutiques.invocationSchemaHandler import InvocationValidationError
from boutiques.localExec import ExecutorError, ExecutorOutput, addDefaultValues
from boutiques.logger import print_error, print_info, raise_error
from boutiques.nexusHelper import NexusError
from boutiques.publisher import ZenodoError
from boutiques.util.utils import (
    customSortInvocationByInput,
    formatSphinxUsage,
    importCatcher,
    loadJson,
)
from boutiques.validator import DescriptorValidationError


def pprint(*params):
    parser = parser_bosh()
    params = ("pprint",) + params
    results = parser.parse_args(params)

    from boutiques.prettyprint import PrettyPrinter

    desc = loadJson(results.descriptor, sandbox=results.sandbox)
    prettyclass = PrettyPrinter(desc)
    return prettyclass.docstring


def create(*params):
    parser = parser_bosh()
    params = ("create",) + params
    results = parser.parse_args(params)

    from boutiques.creator import CreateDescriptor

    new = CreateDescriptor(
        parser=None,
        docker_image=results.docker_image,
        use_singularity=results.use_singularity,
        camel_case=results.camel_case,
        cl_template=results.cl_template,
    )
    new.save(results.descriptor)
    return None


def validate(*params):
    parser = parser_bosh()
    params = ("validate",) + params
    results = parser.parse_args(params)

    from boutiques.validator import validate_descriptor

    descriptor = loadJson(results.descriptor, sandbox=results.sandbox)
    descriptor = validate_descriptor(
        descriptor,
        descriptor_path=results.descriptor,
        format_output=results.format,
        sandbox=results.sandbox,
    )
    if results.bids:
        from boutiques.bids import validate_bids

        validate_bids(descriptor, valid=True)


def execute(*params):
    parser = parser_bosh()
    params = ("exec",) + params
    try:
        # Try to parse input with argparse
        results, _ = parser.parse_known_args(params)
    except SystemExit as e:
        print_info(execute.__doc__)
        raise_error(ExecutorError, "Incorrect usage of 'bosh exec'")

    # Validate mode is in params
    if not hasattr(results, "mode"):
        parser.parse_known_args(params + ("--help",))
        raise_error(ExecutorError, "Missing exec mode {launch, prepare, simulate}.")

    elif results.mode == "launch":
        descriptor = results.descriptor
        inp = results.invocation

        # Validate invocation and descriptor
        arguments = [descriptor, "-i", inp]
        if results.sandbox:
            arguments.append("--sandbox")
        valid = invocation(*arguments)

        # Generate object that will perform the commands
        from boutiques.localExec import LocalExecutor

        executor = LocalExecutor(
            descriptor,
            inp,
            {
                "forcePathType": True,
                "debug": results.debug,
                "changeUser": results.user,
                "stream": results.stream,
                "imagePath": results.imagepath,
                "skipDataCollect": results.skip_data_collection,
                "forceDocker": results.force_docker,
                "forceSingularity": results.force_singularity,
                "provenance": results.provenance,
                "noContainer": results.no_container,
                "sandbox": results.sandbox,
                "noAutomounts": results.no_automounts,
            },
        )
        # Execute it
        return executor.execute(results.volumes, results.container_opts)

    elif results.mode == "simulate":
        descriptor = results.descriptor

        # Do some basic input scrubbing
        inp = results.input

        arguments = [descriptor]
        if inp:
            arguments.append("-i")
            arguments.append(inp)
        if results.sandbox:
            arguments.append("--sandbox")
        valid = invocation(*arguments)

        # Generate object that will perform the commands
        from boutiques.localExec import LocalExecutor

        executor = LocalExecutor(
            descriptor,
            inp,
            {
                "forcePathType": True,
                "destroyTempScripts": True,
                "changeUser": True,
                "skipDataCollect": True,
                "requireComplete": results.complete,
                "sandbox": results.sandbox,
            },
        )
        if not inp:
            # Add optional inputs with default-value to inputs_dict,
            # which is then populated with random params
            executor.in_dict = addDefaultValues(executor.desc_dict, {})
            executor.generateRandomParams(generateCmdLineFromInDict=True)

        if results.json:
            sout = [
                json.dumps(
                    customSortInvocationByInput(executor.in_dict, descriptor),
                    indent=4,
                )
            ]
            print(sout[0])
        else:
            executor.printCmdLine()
            sout = executor.cmd_line

        # for consistency with execute
        # Adding hide to "container location" field since it's an invalid
        # value, can parse that to hide the summary print
        return ExecutorOutput(
            os.linesep.join(sout),
            "",
            0,
            "",
            [],
            [],
            os.linesep.join(sout),
            "",
            "hide",
        )

    elif results.mode == "prepare":
        descriptor = results.descriptor

        # Validate descriptor
        arguments = [descriptor]
        if results.sandbox:
            arguments.append("--sandbox")
        valid = invocation(*arguments)

        # Generate object that will perform the commands
        from boutiques.localExec import LocalExecutor

        executor = LocalExecutor(
            descriptor,
            None,
            {
                "forcePathType": True,
                "debug": results.debug,
                "stream": results.stream,
                "imagePath": results.imagepath,
                "skipDataCollect": True,
                "sandbox": results.sandbox,
            },
        )
        container_location = executor.prepare()[1]
        print("Container location: " + container_location)

        # Adding hide to "container location" field since it's an invalid
        # value, and we can parse that to hide the summary print
        return ExecutorOutput(container_location, "", 0, "", [], [], "", "", "hide")


def importer(*params):
    parser = parser_bosh()
    params = ("import",) + params
    results = parser.parse_args(params)

    from boutiques.importer import Importer

    importer = Importer(
        results.input_descriptor,
        results.output_descriptor,
        results.input_invocation,
        results.output_invocation,
    )
    if results.type == "0.4":
        importer.upgrade_04()
    elif results.type == "bids":
        importer.import_bids()
    elif results.type == "cwl":
        importer.import_cwl()
    elif results.type == "docopt":
        create(results.output_descriptor)
        importer.import_docopt()
    elif results.type == "config":
        create(results.output_descriptor)
        importer.import_config()


def exporter(*params):
    parser = parser_bosh()
    params = ("export",) + params
    results = parser.parse_args(params)

    descriptor = results.descriptor
    output = results.output

    args = [results.descriptor]
    if results.sandbox:
        args.append("--sandbox")
    validate(*args)

    from boutiques.exporter import Exporter

    exporter = Exporter(descriptor, results.identifier, sandbox=results.sandbox)
    if results.type == "carmin":
        exporter.carmin(output)


def publish(*params):
    parser = parser_bosh()
    params = ("publish",) + params
    results = parser.parse_args(params)

    from boutiques.publisher import Publisher

    publisher = Publisher(
        results.boutiques_descriptor,
        results.zenodo_token,
        results.verbose,
        results.sandbox,
        results.no_int,
        results.replace,
        results.id,
    )
    publisher.publish()
    if hasattr(publisher, "doi"):
        return publisher.doi


def invocation(*params):
    parser = parser_bosh()
    params = ("invocation",) + params
    results = parser.parse_args(params)
    arguments = [results.descriptor]
    if results.sandbox:
        arguments.append("--sandbox")
    validate(*arguments)
    descriptor = loadJson(results.descriptor, sandbox=results.sandbox)
    if descriptor.get("invocation-schema"):
        invSchema = descriptor.get("invocation-schema")
    else:
        from boutiques.invocationSchemaHandler import generateInvocationSchema

        invSchema = generateInvocationSchema(descriptor)
        if results.write_schema:
            descriptor["invocation-schema"] = invSchema
            with open(results.descriptor, "w") as f:
                f.write(json.dumps(descriptor, indent=4))
    if results.invocation:
        from boutiques.invocationSchemaHandler import validateSchema

        data = addDefaultValues(descriptor, loadJson(results.invocation))
        validateSchema(invSchema, data)


def evaluate(*params):
    parser = parser_bosh()
    params = ("evaluate",) + params
    results = parser.parse_args(params)

    # Generate object that will parse the invocation and descriptor
    from boutiques.localExec import LocalExecutor

    executor = LocalExecutor(
        results.descriptor,
        results.invocation,
        {
            "forcePathType": True,
            "destroyTempScripts": True,
            "changeUser": True,
            "skipDataCollect": True,
            "sandbox": results.sandbox,
        },
    )

    from boutiques.evaluate import evaluateEngine

    query_results = []
    for query in results.query:
        query_results += [evaluateEngine(executor, query)]
    return query_results[0] if len(query_results) == 1 else query_results


@importCatcher()
def test(*params):
    parser = parser_bosh()
    params = ("test",) + params
    results = parser.parse_args(params)

    args = [results.descriptor]
    if results.sandbox:
        args.append("--sandbox")

    # Generation of the invocation schema (and descriptor validation).
    invocation(*args)

    # Extraction of all the invocations defined for the test-cases.
    descriptor = loadJson(results.descriptor, sandbox=results.sandbox)

    if not descriptor.get("tests"):
        # If no tests have been specified, we consider testing successful.
        return 0

    for test in descriptor["tests"]:
        invocation_JSON = test["invocation"]

        testArgs = [
            results.descriptor,
            "--invocation",
            json.dumps(invocation_JSON),
        ]
        if results.sandbox:
            testArgs.append("--sandbox")
        # Check if the invocation is valid.
        invocation(*testArgs)

    # Invocations have been properly validated. We can launch the actual tests.
    test_path = op.join(op.dirname(op.realpath(__file__)), "test.py")
    test_args = [test_path, "--descriptor", results.descriptor]
    if results.imagepath:
        test_args.extend(["--imagepath", results.imagepath])
    import pytest

    return pytest.main(args=test_args)


def search(*params):
    parser = parser_bosh()
    params = ("search",) + params
    results = parser.parse_args(params)

    from boutiques.searcher import Searcher

    searcher = Searcher(
        results.query,
        results.verbose,
        results.sandbox,
        results.max,
        results.no_trunc,
        results.exact,
    )

    return searcher.search()


def example(*params):
    parser = parser_bosh()
    params = ("example",) + params
    results = parser.parse_args(params)

    descriptor = results.descriptor
    arguments = [descriptor]
    if results.sandbox:
        arguments.append("--sandbox")
    valid = invocation(*arguments)

    # Generate object that will perform the commands
    from boutiques.localExec import LocalExecutor

    executor = LocalExecutor(
        descriptor,
        None,
        {
            "forcePathType": True,
            "destroyTempScripts": True,
            "changeUser": True,
            "skipDataCollect": True,
            "requireComplete": results.complete,
            "sandbox": results.sandbox,
        },
    )
    executor.generateRandomParams()
    return json.dumps(
        customSortInvocationByInput(executor.in_dict, descriptor), indent=4
    )


def pull(*params):
    parser = parser_bosh()
    params = ("pull",) + params
    results = parser.parse_args(params)

    from boutiques.puller import Puller

    puller = Puller(results.zids, results.verbose, results.sandbox)
    return puller.pull()


def data(*params):
    parser = parser_bosh()
    params = ("data",) + params
    try:
        # Try to parse input with argparse
        results, _ = parser.parse_known_args(params)
    except SystemExit as e:
        print_info(data.__doc__)
        raise_error(DataHandlerError, "Incorrect usage of 'bosh data'")

    # Validate mode is in params
    if not hasattr(results, "mode"):
        parser.parse_known_args(params + ("--help",))
        raise_error(DataHandlerError, "Missing data mode {delete, inspect, publish}.")
    elif results.mode == "inspect":
        from boutiques.dataHandler import DataHandler

        dataHandler = DataHandler()
        return dataHandler.inspect(results.example, results.latest)
    elif results.mode == "publish":
        from boutiques.dataHandler import DataHandler

        dataHandler = DataHandler()
        return dataHandler.publish(
            results.file,
            results.zenodo_token,
            results.author,
            results.nexus_token,
            results.nexus_org,
            results.nexus_project,
            results.individually,
            results.sandbox,
            results.no_int,
            results.verbose,
            results.nexus,
        )
    elif results.mode == "delete":
        from boutiques.dataHandler import DataHandler

        dataHandler = DataHandler()
        return dataHandler.delete(results.file, results.no_int)


def deprecate(*params):
    parser = parser_bosh()
    params = ("deprecate",) + params
    result = parser.parse_args(params)

    from boutiques.deprecate import deprecate

    return deprecate(
        result.zid,
        by_zenodo_id=result.by,
        sandbox=result.sandbox,
        verbose=result.verbose,
        zenodo_token=result.zenodo_token,
    )


def bosh(args=None):
    # Returns True if bosh was called from the CLI
    def runs_as_cli():
        return os.path.basename(sys.argv[0]) == "bosh"

    def bosh_return(val, code=0, hide=False, formatted=None):
        if runs_as_cli():
            if not hide and val is not None:
                print(formatted if formatted is not None else val)
            elif not hide:
                print("OK" if code == 0 else "Failed")
            return code  # everything went well
        return val  # calling function wants this value

    # Params are set depending on where bosh is called from
    if runs_as_cli():
        func = sys.argv[1] if len(sys.argv) >= 2 else None
        params = sys.argv[2:] if func is not None else []
    else:
        func = args[0] if len(args) > 0 else args
        params = args[1:] if func is not None else []

    try:
        if func == "create":
            out = create(*params)
            return bosh_return(out, hide=True)
        elif func == "validate":
            out = validate(*params)
            return bosh_return(out)
        elif func in ["exec", "execute"]:
            out = execute(*params)
            # If executed through CLI, print 'out' and return exit_code
            # Otherwise, return out
            return bosh_return(
                out, out.exit_code, hide=bool(out.container_location == "hide")
            )
        elif func == "example":
            out = example(*params)
            return bosh_return(out)
        elif func == "import":
            out = importer(*params)
            return bosh_return(out)
        elif func == "export":
            out = exporter(*params)
            return bosh_return(out)
        elif func == "publish":
            out = publish(*params)
            return bosh_return(out)
        elif func == "invocation":
            out = invocation(*params)
            return bosh_return(out)
        elif func == "evaluate":
            out = evaluate(*params)
            return bosh_return(out)
        elif func == "test":
            out = test(*params)
            return bosh_return(out)
        elif func == "pprint":
            out = pprint(*params)
            return bosh_return(out)
        elif func == "search":
            out = search(*params)
            return bosh_return(
                out, formatted=tabulate(out, headers="keys", tablefmt="plain")
            )
        elif func == "pull":
            out = pull(*params)
            return bosh_return(out, hide=True)
        elif func == "data":
            out = data(*params)
            return bosh_return(out)
        elif func == "version":
            from boutiques.__version__ import VERSION

            return bosh_return(VERSION)
        elif func == "deprecate":
            out = deprecate(*params)
            return bosh_return(out)
        else:
            print(parser_bosh().format_help())
            raise_error(ExecutorError, f"Incorrect bosh mode '{func}'")

    except (
        ZenodoError,
        NexusError,
        DescriptorValidationError,
        InvocationValidationError,
        ValidationError,
        ExportError,
        ImportError,
        ExecutorError,
    ) as e:
        # We don't want to raise an exception when function is called
        # from CLI.'
        if runs_as_cli():
            print(e)
            return 99  # Note: this conflicts with tool error codes.
        raise e
    except SystemExit as e:
        if runs_as_cli():
            print(e)
            return 99  # Note: this conflicts with tool error codes.
        raise_error(
            BoutiquesError,
            "Unable to parse arguments resulting in SystemExit.",
        )


class BoutiquesError(Exception):
    pass


# This section is for documentation generation purposes
bosh.__doc__ = parser_bosh().format_usage().replace("sphinx-build", "bosh")
# retrieve subparsers from parser
subparsers_actions = [
    a for a in parser_bosh()._actions if isinstance(a, argparse._SubParsersAction)
]
for action in subparsers_actions:
    # get all subparsers and assign __doc__ to functions
    for func, subparser in action.choices.items():
        usage_string = formatSphinxUsage(func, subparser.format_usage())
        if func == "create":
            create.__doc__ = usage_string
        elif func == "data":
            data.__doc__ = usage_string
        elif func == "deprecate":
            deprecate.__doc__ = usage_string
        elif func == "evaluate":
            evaluate.__doc__ = usage_string
        elif func == "example":
            example.__doc__ = usage_string
        elif func == "exec":
            execute.__doc__ = usage_string
        elif func == "export":
            exporter.__doc__ = usage_string
        elif func == "import":
            importer.__doc__ = usage_string
        elif func == "invocation":
            invocation.__doc__ = usage_string
        elif func == "pprint":
            pprint.__doc__ = usage_string
        elif func == "publish":
            publish.__doc__ = usage_string
        elif func == "pull":
            pull.__doc__ = usage_string
        elif func == "search":
            search.__doc__ = usage_string
        elif func == "test":
            test.__doc__ = usage_string
        elif func == "validate":
            validate.__doc__ = usage_string



#!/usr/bin/env python
import os
import re

import simplejson as json

from boutiques.logger import print_info, raise_error
from boutiques.util.utils import importCatcher


class ZenodoError(Exception):
    pass


class ZenodoHelper:

    # Constructor
    def __init__(self, sandbox=False, no_int=False, verbose=False):
        self.sandbox = sandbox
        self.no_int = no_int
        self.verbose = verbose
        self.config_file = os.path.join(os.path.expanduser("~"), ".boutiques")
        self.zenodo_endpoint = self.get_zenodo_endpoint()

    def verify_zenodo_access_token(self, user_input):
        access_token = user_input
        if access_token is None:
            access_token = self.get_zenodo_access_token()
        self.zenodo_test_api(access_token)
        self.save_zenodo_access_token(access_token)
        return access_token

    def get_zenodo_access_token(self):
        json_creds = self.read_credentials()
        if json_creds.get(self.config_token_property_name()):
            return json_creds.get(self.config_token_property_name())
        if self.no_int:
            raise_error(ZenodoError, "Cannot find Zenodo credentials.")
        prompt = (
            "Please enter your Zenodo access token (it will be "
            "saved in {} for future use): ".format(self.config_file)
        )
        try:
            return raw_input(prompt)  # Python 2
        except NameError:
            return input(prompt)  # Python 3

    def save_zenodo_access_token(self, access_token):
        json_creds = self.read_credentials()
        json_creds[self.config_token_property_name()] = access_token
        with open(self.config_file, "w") as f:
            f.write(json.dumps(json_creds, indent=4, sort_keys=True))
        if self.verbose:
            print_info(f"Zenodo access token saved in {self.config_file}")

    def get_zenodo_endpoint(self):
        endpoint = (
            "https://sandbox.zenodo.org" if self.sandbox else "https://zenodo.org"
        )
        if self.verbose:
            print_info(f"Using Zenodo endpoint {endpoint}")
        return endpoint

    @importCatcher()
    def record_exists(self, record_id):
        import requests

        r = requests.get(self.zenodo_endpoint + f"/api/records/{record_id}")
        if r.status_code == 200:
            return True
        if r.status_code == 404:
            return False
        raise_error(ZenodoError, f"Cannot test existence of record {record_id}", r)

    @importCatcher()
    def zenodo_get_record(self, zenodo_id):
        import requests

        r = requests.get(self.zenodo_endpoint + f"/api/records/{zenodo_id}")
        if r.status_code != 200:
            raise_error(ZenodoError, f'Descriptor "{zenodo_id}" not found', r)
        return r.json()

    def get_record_id_from_zid(self, zenodo_id):
        """
        zenodo_id is in the form zenodo.1234567
        record id is 1234567
        """
        if not re.match(r"zenodo\.[0-9]", zenodo_id):
            raise_error(
                ZenodoError,
                "This does not look like a valid Zenodo ID: {}."
                "Zenodo ids must be in the form zenodo.1234567".format(zenodo_id),
            )
        parts = zenodo_id.split(".")
        return parts[1]

    def get_zid_from_filename(self, filename):
        # Filename must be in the form /a/b/c/zenodo-1234.json
        # where zenodo.1234 is the record id.
        basename = os.path.basename(filename)
        if not re.match(r"zenodo-[0-9]*\.json", basename):
            raise_error(
                ZenodoError,
                f"This does not look like a valid file name: {filename}",
            )
        return basename.replace(".json", "").replace("-", ".")

    def get_doi_from_zid(self, zenodo_id):
        prefix = "10.5072" if self.sandbox else "10.5281"
        return f"{prefix}/{zenodo_id}"

    def config_token_property_name(self):
        if self.sandbox:
            return "zenodo-access-token-test"
        return "zenodo-access-token"

    def read_credentials(self):
        try:
            with open(self.config_file) as f:
                json_creds = json.load(f)
        except OSError:
            json_creds = {}
        except ValueError:
            json_creds = {}
        return json_creds

    @importCatcher()
    def zenodo_test_api(self, access_token):
        import requests

        r = requests.get(
            self.zenodo_endpoint + "/api/deposit/depositions",
        )
        if r.status_code != 403:
            raise_error(
                ZenodoError, f"Cannot access Zenodo at {self.zenodo_endpoint}", r
            )
        if self.verbose:
            print_info("Zenodo is accessible", r)

        r = requests.get(
            self.zenodo_endpoint + "/api/deposit/depositions",
            params={"access_token": access_token},
        )
        message = "Cannot authenticate to Zenodo API, check your access token"
        if r.status_code != 200:
            raise_error(ZenodoError, message, r)
        if self.verbose:
            print_info("Authentication to Zenodo successful", r)

    @importCatcher()
    def zenodo_deposit(self, metadata, access_token):
        import requests

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        }
        data = metadata
        r = requests.post(
            self.zenodo_endpoint + "/api/deposit/depositions",
            json={},
            data=json.dumps(data),
            headers=headers,
        )
        if r.status_code != 201:
            raise_error(ZenodoError, "Deposition failed", r)
        zid = r.json()["id"]
        if self.verbose:
            print_info(f"Deposition succeeded, id is {zid}", r)
        return zid

    @importCatcher()
    def zenodo_deposit_updated_version(self, metadata, access_token, deposition_id):
        import requests

        r = requests.post(
            self.zenodo_endpoint
            + f"/api/deposit/depositions/{deposition_id}/actions/newversion",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        if r.status_code == 403:
            raise_error(
                ZenodoError,
                "You do not have permission to access "
                "this resource. Note that you cannot "
                "publish an update to a tool belonging "
                "to someone else.",
                r,
            )
        elif r.status_code != 201:
            raise_error(
                ZenodoError,
                "Deposition of new version failed. Check "
                "that the Zenodo ID is correct (if one "
                "was provided).",
                r,
            )
        if self.verbose:
            print_info("Deposition of new version succeeded", r)
        new_url = r.json()["links"]["latest_draft"]
        new_zid = new_url.split("/")[-1]
        self.zenodo_update_metadata(new_zid, metadata, access_token)
        self.zenodo_delete_files(new_zid, r.json()["files"], access_token)
        return new_zid

    @importCatcher()
    def zenodo_update_metadata(self, new_deposition_id, metadata, access_token):
        import requests

        data = metadata

        headers = {
            "Content-Type": "application/json",
            "Authorization": f"Bearer {access_token}",
        }
        r = requests.put(
            self.zenodo_endpoint + "/api/deposit/depositions/%s" % new_deposition_id,
            data=json.dumps(data),
            headers=headers,
        )
        if r.status_code != 200:
            raise_error(ZenodoError, "Cannot update metadata of new version", r)
        if self.verbose:
            print_info("Updated metadata of new version", r)

    # When a new version is created, the files from the old version are
    # automatically copied over. This method removes them.
    @importCatcher()
    def zenodo_delete_files(self, new_deposition_id, files, access_token):
        import requests

        for file in files:
            file_id = file["id"]
            r = requests.delete(
                self.zenodo_endpoint
                + f"/api/deposit/depositions/{new_deposition_id}/files/{file_id}",
                headers={"Authorization": f"Bearer {access_token}"},
            )
            if r.status_code != 204:
                raise_error(ZenodoError, "Could not delete old file", r)
            if self.verbose:
                print_info("Deleted old file", r)

    @importCatcher()
    def zenodo_publish(self, access_token, deposition_id, msg_obj):
        import requests

        r = requests.post(
            self.zenodo_endpoint
            + f"/api/deposit/depositions/{deposition_id}/actions/publish",
            headers={"Authorization": f"Bearer {access_token}"},
        )
        if r.status_code != 202:
            raise_error(ZenodoError, f"Cannot publish {msg_obj}", r)
        if self.verbose:
            print_info(
                f"{msg_obj} published to Zenodo, doi is {r.json()['doi']}",
                r,
            )
        return r.json()["doi"]

    @importCatcher()
    def zenodo_search(self, query, query_line):
        import requests

        # Get all results
        get_request = self.zenodo_endpoint + (
            "/api/records/?q="
            "keywords:(/Boutiques/) AND "
            "keywords:(/schema.*/) AND keywords:(/version.*/)"
            "%s"
            "&file_type=json&type=software&"
            "page=1&size=%s" % (query_line, 9999)
        )
        r = requests.get(get_request)
        if r.status_code != 200:
            raise_error(ZenodoError, "Error searching Zenodo", r)
        if self.verbose:
            print_info(f'Search successful for query "{query}"', r)
            print_info(f"GET request: {get_request}")
        return r

    @importCatcher()
    def zenodo_upload_file(
        self,
        deposition_id,
        file_path,
        zenodo_access_token=None,
        error_msg="Cannot Upload to Zenodo",
        verbose_msg="Uploaded to Zenodo",
    ):
        import requests

        zenodo_access_token = (
            self.get_zenodo_access_token
            if zenodo_access_token is None
            else zenodo_access_token
        )
        r = requests.post(
            self.zenodo_endpoint + f"/api/deposit/depositions/{deposition_id}/files",
            headers={"Authorization": f"Bearer {zenodo_access_token}"},
            data={"filename": os.path.basename(file_path)},
            files={"file": open(file_path, "rb")},
        )

        if r.status_code != 201:
            raise_error(ZenodoError, error_msg, r)
        if self.verbose:
            print_info(verbose_msg, r)



#!/usr/bin/env python

import pytest

import boutiques as bosh
from boutiques.tests.BaseTest import BaseTest


class TestExampleFlag(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("example_flag")

    def test_example_flag_1(self):
        ret = bosh.execute(
            "simulate",
            self.get_file_path("example-flag.json"),
            "-i",
            self.get_file_path("i1.json"),
        )
        self.assertEqual(ret.shell_command.strip(), "/bin/true -a -b")

    def test_example_flag_2(self):
        ret = bosh.execute(
            "simulate",
            self.get_file_path("example-flag.json"),
            "-i",
            self.get_file_path("i2.json"),
        )
        self.assertEqual(ret.shell_command.strip(), "/bin/true")

    def test_example_flag_3(self):
        ret = bosh.execute(
            "simulate",
            self.get_file_path("example-flag.json"),
            "-i",
            self.get_file_path("i3.json"),
        )

        self.assertEqual(ret.shell_command.replace("  ", " ").strip(), "/bin/true -b")



#!/usr/bin/env python
import hashlib
import os
import time

from boutiques.logger import print_info, raise_error
from boutiques.nexusHelper import NexusHelper
from boutiques.util.utils import extractFileName, importCatcher, loadJson


class DataHandler:

    # Constructor
    def __init__(self):
        self.cache_dir = getDataCacheDir()
        self.cache_files = os.listdir(self.cache_dir)
        self.descriptor_files = [
            fl for fl in self.cache_files if fl.split("_")[0] == "descriptor"
        ]
        self.record_files = [
            fl for fl in self.cache_files if fl not in self.descriptor_files
        ]

    # Function to display the contents of the cache
    # Option to display an example file which displays an pre-made example file
    # or the first file of the cache if it exists
    # Otherwise displays information about the cache and a list of files
    def inspect(self, example=False, latest=False):
        self.example = example
        self.latest = latest
        # Display an example json file
        if self.example:
            # Display the first file in cache
            if len(self.record_files) > 0:
                filename = self.record_files[0]
                file_path = os.path.join(self.cache_dir, filename)
                self._display_file(file_path)
            else:
                print("No records in the cache at the moment.")
        elif self.latest:
            if len(self.record_files) > 0:
                self.record_files.sort(
                    key=lambda x: os.path.getmtime(os.path.join(self.cache_dir, x)),
                    reverse=True,
                )
                filename = self.record_files[0]
                file_path = os.path.join(self.cache_dir, filename)
                self._display_file(file_path)
            else:
                print("No records in the cache at the moment.")
        # Print information about files in cache
        else:
            print(
                f"There are {len(self.record_files)} unpublished records in the cache"
            )
            print(
                "There are {} unpublished descriptors in the cache".format(
                    len(self.descriptor_files)
                )
            )
            for i in range(len(self.cache_files)):
                print(self.cache_files[i])

    # Private function to print a file to console
    def _display_file(self, file_path):
        with open(file_path) as file_in:
            print(file_in.read())
            file_in.close()

    # Function to publish a data set to Zenodo or Nexus
    # Options allow to only publish a single file, publish files individually as
    # data sets or bulk publish all files in the cache as one data set (default)
    @importCatcher()
    def publish(
        self,
        file,
        zenodo_token,
        author,
        nexus_token,
        nexus_org,
        nexus_project,
        individually=False,
        sandbox=False,
        no_int=False,
        verbose=False,
        to_nexus=False,
    ):
        self.filename = extractFileName(file)
        self.author = "Anonymous"
        if author is not None:
            self.author = author
        self.individual = individually
        self.sandbox = sandbox
        self.no_int = no_int
        self.verbose = verbose
        self.to_nexus = to_nexus
        if not self.to_nexus:
            from boutiques.zenodoHelper import ZenodoHelper

            self.zenodo_access_token = zenodo_token
            self.zenodo_helper = ZenodoHelper(sandbox, no_int, verbose)
            self.zenodo_endpoint = self.zenodo_helper.zenodo_endpoint
            self.zenodo_access_token = self.zenodo_helper.verify_zenodo_access_token(
                self.zenodo_access_token
            )
        else:
            self.nexus_helper = NexusHelper(sandbox, no_int, verbose)
            self.nexus_endpoint = self.nexus_helper.nexus_endpoint
            self.nexus_access_token, self.nexus_org, self.nexus_project = (
                self.nexus_helper.verify_nexus_input(
                    nexus_token, nexus_org, nexus_project
                )
            )

        # Verify publishing
        if not self.no_int:
            prompt = self._get_publishing_prompt()
            try:
                ret = raw_input(prompt)  # Python 2
            except NameError:
                ret = input(prompt)  # Python 3
            if ret.upper() != "Y":
                return

        # Flag for data-set size
        self.bulk_publish = False
        # Single record publication
        if self.filename is not None:
            self._file_exists_in_cache(self.filename)
            self._publish([self.filename])
        # All records published to individual data-sets
        elif individually:
            for file in self.record_files:
                self._publish([file])
        # All records published to one data-set
        else:
            self.bulk_publish = True
            self._publish(self.record_files)

    # Private method to publish all the records in file_list to a single
    # data-set on Zenodo or Nexus
    def _publish(self, files_list):
        # Filter files_list by records with Zenodo ids
        records_dict = self._checkPublishable(files_list)
        # Publishable list is empty, end execution
        if len(records_dict) == 0:
            return

        # Publish to Nexus
        if self.to_nexus:
            for file in files_list:
                self.nexus_helper.publish(
                    self.nexus_org,
                    self.nexus_project,
                    os.path.join(self.cache_dir, file),
                )

        # Publish to Zenodo
        else:
            # Create deposition
            deposition_id = self.zenodo_helper.zenodo_deposit(
                self._create_metadata(records_dict), self.zenodo_access_token
            )

            # Upload all files in files_list to deposition
            for fil in records_dict.keys():
                file_path = os.path.join(self.cache_dir, fil)
                self.zenodo_helper.zenodo_upload_file(
                    deposition_id,
                    file_path,
                    zenodo_access_token=self.zenodo_access_token,
                    error_msg="Cannot upload record to Zenodo",
                    verbose_msg="Record uploaded to Zenodo",
                )

            # Publish deposition
            msg_obj = "Records" if self.bulk_publish else "Record"
            doi = self.zenodo_helper.zenodo_publish(
                self.zenodo_access_token, deposition_id, msg_obj
            )
            # Clear cache of published records
            if doi:
                self._clean_cache(records_dict)

    # Private function to filter out records that can not be published
    # because they lack a descriptor doi
    def _checkPublishable(self, files_list):
        # dict {filename: content_dict}
        desc_to_publish = set()
        publishable_dict = {}
        for fl in files_list:
            fl_path = os.path.join(self.cache_dir, fl)
            fl_dict = loadJson(fl_path)
            doi = fl_dict.get("summary").get("descriptor-doi")
            # Descriptor is not publish, record contains link to file
            if doi.split("_")[0] == "descriptor":
                desc_path = os.path.join(self.cache_dir, doi)
                desc_dict = loadJson(desc_path)
                # Descriptor is published, record needs to be updated
                if desc_dict.get("doi") is not None:
                    fl_dict["summary"]["descriptor-doi"] = desc_dict["doi"]
                    publishable_dict[fl] = fl_dict
                # Descriptor isn't published, inform user with full prompt
                else:
                    print(
                        "Record {} cannot be published as its descriptor "
                        "is not yet published. ".format(fl)
                    )
                    desc_to_publish.add(f"bosh publish {desc_path}")
            # Descriptor doi is stored correctly in record
            else:
                publishable_dict[fl] = fl_dict
        # Prompt user to publish descriptors
        if len(desc_to_publish) != 0:
            print(
                "Some descriptors have not been published, they can be "
                "published with following commands:"
            )
            for prompt in desc_to_publish:
                print("\t" + prompt)
        return publishable_dict

    def _create_metadata(self, records_dict):
        url = "https://zenodo.org/record/{}"
        hash = hashlib.md5()
        hash.update(str(time.time()).encode("utf-8"))
        identifier = hash.hexdigest()
        data = {
            "metadata": {
                "title": f"Boutiques-execution-{identifier[:6]}",
                "upload_type": "dataset",
                "description": "Boutiques execution data-set",
                "creators": [{"name": self.author}],
            }
        }

        # Get unique list of tool names and descriptors
        unique_names = {v["summary"]["name"] for v in records_dict.values()}
        unique_descriptors = {
            v["summary"]["descriptor-doi"] for v in records_dict.values()
        }

        # Add tool name(s) to keywords
        data["metadata"]["keywords"] = [v for v in unique_names]
        data["metadata"]["keywords"].insert(0, "Boutiques")
        # Add descriptor link(s) to related identifiers
        data["metadata"]["related_identifiers"] = [
            {"identifier": url.format(v.split(".")[2]), "relation": "hasPart"}
            for v in unique_descriptors
        ]
        return data

    def _get_publishing_prompt(self):
        _destination = (
            "Nexus in organization '{}' and project '{}'".format(
                self.nexus_org, self.nexus_project
            )
            if self.to_nexus
            else "Zenodo"
        )
        if self.filename is not None:
            return (
                "The dataset {} will be published to {}, "
                "this cannot be undone. Are you sure? (Y/n) ".format(
                    self.filename, _destination
                )
            )
        if self.individual:
            return (
                "The records will be published to {} each as "
                "separate data-sets. This cannot be undone. Are you "
                "sure? (Y/n ".format(_destination)
            )
        return (
            "The records will be published to {} as a data-set. This "
            "cannot be undone. Are you sure? (Y/n) ".format(_destination)
        )

    # Private function to remove published files and descriptors which no
    # longer have dependencies
    def _clean_cache(self, records_dict):
        for record in records_dict.keys():
            self.delete(record, True)
        # List remaining records and collect descriptor-doi values
        self.record_files = [
            fl for fl in os.listdir(self.cache_dir) if fl not in self.descriptor_files
        ]
        doi_list = [
            loadJson(os.path.join(self.cache_dir, fl))
            .get("summary")
            .get("descriptor-doi")
            for fl in self.record_files
        ]

        # Check each descriptor in remaining records
        for descriptor in self.descriptor_files:
            # No records link to descriptor
            if descriptor not in doi_list:
                self.delete(descriptor, True)
                self.descriptor_files.remove(descriptor)

    # Function to remove file(s) from the cache
    # Option all will clear the data collection cache of all files
    # or passing in a file will remove only that file
    # Options are mutually exclusive and one is required
    def delete(self, file=None, no_int=False):
        self.filename = extractFileName(file)
        self.no_int = no_int

        # Verify deletion
        if not self.no_int:
            prompt = self._get_delete_prompt()
            try:
                ret = raw_input(prompt)  # Python 2
            except NameError:
                ret = input(prompt)  # Python 3
            if ret.upper() != "Y":
                return

        # Remove the file specified by the file option
        if file is not None:
            # Check file exists in cache
            self._file_exists_in_cache(file)
            # Remove file from cache
            file_path = os.path.join(self.cache_dir, file)
            os.remove(file_path)
            print_info(f"File {file} has been removed from the data cache")
        # Remove all files in the data cache
        else:
            [os.remove(os.path.join(self.cache_dir, f)) for f in self.cache_files]
            print_info("All files have been removed from the data cache")

    def _file_exists_in_cache(self, filename):
        file_path = os.path.join(self.cache_dir, filename)
        # Incorrect filename input
        if not os.path.isfile(file_path):
            msg = f"File {filename} does not exist in the data cache"
            raise_error(ValueError, msg)

    def _get_delete_prompt(self):
        if self.filename is not None:
            return (
                "The dataset {} will be deleted from the cache, "
                "this cannot be undone. Are you sure? (Y/n) ".format(self.filename)
            )
        return (
            "All records will be removed from the cache. This "
            "cannot be undone. Are you sure? (Y/n) "
        )


def getDataCacheDir():
    cache_dir = os.path.join(os.path.expanduser("~"), ".cache", "boutiques")
    data_cache_dir = os.path.join(cache_dir, "data")
    if not os.path.exists(cache_dir):
        os.makedirs(cache_dir, exist_ok=True)
    if not os.path.exists(data_cache_dir):
        os.makedirs(data_cache_dir, exist_ok=True)
    return data_cache_dir


class DataHandlerError(Exception):
    pass



#!/usr/bin/env python

import subprocess

import pytest
import simplejson as json

from boutiques.tests.BaseTest import BaseTest


class TestExample(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("example")

    def test_example_complete(self):
        descriptor = self.get_file_path("test_example_descriptor.json")
        command = "bosh example " + descriptor + " -c"
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        output = json.loads(process.stdout.read())

        self.assertTrue(("a1" in output and "b2" in output) or "a2" in output)
        self.assertIn("b1", output)
        self.assertIn("c1", output)
        self.assertIn("c2", output)
        self.assertIn("d1", output)
        self.assertIn("d2", output)

    def test_example_required_only(self):
        descriptor = self.get_file_path("test_example_descriptor.json")
        command = "bosh example " + descriptor
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        output = json.loads(process.stdout.read())

        self.assertDictEqual({"b1": "b1", "c2": "c2"}, output)

    def test_example_requires_group_complete_x10(self):
        descriptor = self.get_file_path("test_docopt_valid.json")
        from boutiques.localExec import LocalExecutor

        executor = LocalExecutor(
            descriptor,
            None,
            {
                "forcePathType": True,
                "destroyTempScripts": True,
                "changeUser": True,
                "skipDataCollect": True,
                "requireComplete": True,
                "sandbox": False,
            },
        )

        # Can't create descriptors with mutex group but only one valid example
        # Bosh example is inherently random,
        # Couldn't even inject prederemined input to executor.in_dict
        # because _randomFillInDict clears it
        for _ in range(0, 10):
            executor.generateRandomParams()
            self.assertGreater(len(executor.in_dict), 0)
            executor.in_dict = None



#!/usr/bin/env python
from boutiques import BoutiquesError
from boutiques.bosh import bosh
from boutiques.dataHandler import DataHandlerError
from boutiques.localExec import ExecutorError
from boutiques.tests.BaseTest import BaseTest


class TestBosh(BaseTest):

    def test_help(self):
        self.assertRaises(ExecutorError, bosh, [])
        self.assertRaises(ExecutorError, bosh, ["--help"])
        self.assertRaises(ExecutorError, bosh, ["exec", "--help"])
        self.assertRaises(ExecutorError, bosh, ["exec", "launch", "--help"])
        self.assertRaises(ExecutorError, bosh, ["exec", "simulate", "--help"])
        self.assertRaises(ExecutorError, bosh, ["exec", "prepare", "--help"])
        self.assertRaises(DataHandlerError, bosh, ["data", "--help"])
        self.assertRaises(DataHandlerError, bosh, ["data", "delete", "--help"])
        self.assertRaises(DataHandlerError, bosh, ["data", "inspect", "--help"])
        self.assertRaises(DataHandlerError, bosh, ["data", "publish", "--help"])
        self.assertRaises(BoutiquesError, bosh, ["publish", "--help"])
        self.assertRaises(BoutiquesError, bosh, ["import", "--help"])
        self.assertRaises(BoutiquesError, bosh, ["validate", "--help"])
        self.assertRaises(BoutiquesError, bosh, ["invocation", "--help"])
        self.assertRaises(BoutiquesError, bosh, ["evaluate", "--help"])
        self.assertRaises(BoutiquesError, bosh, ["create", "--help"])
        self.assertRaises(BoutiquesError, bosh, ["example", "--help"])



#!/usr/bin/env python

import os
import os.path as op
import subprocess
import tarfile
from os.path import join as opj

import pytest
import simplejson as json
from jsonschema.exceptions import ValidationError

import boutiques
from boutiques import bosh
from boutiques.importer import ImportError
from boutiques.tests.BaseTest import BaseTest
from boutiques.util.utils import loadJson


class TestImport(BaseTest):
    @pytest.fixture(scope="session", autouse=True)
    def clean_up(self):
        yield
        remove_list = ["user-image.simg", "example.conf", "stdout.txt"]
        for item in remove_list:
            if os.path.isfile(item):
                os.remove(item)

    def test_import_bids_good(self):
        self.setup("bids")
        bids_app = self.get_file_path("example_good")
        outfile = "./test_temp/test-import.json"
        ref_name = "test-import-ref.json"
        if op.isfile(outfile):
            os.remove(outfile)
        self.assertFalse(bosh(["import", "bids", outfile, bids_app]))
        self.assertEqual(
            open(outfile).read().strip(),
            open(opj(bids_app, ref_name)).read().strip(),
        )
        self.assertFalse(bosh(["validate", outfile, "-b"]))

    def test_import_bids_bad(self):
        self.setup("bids")
        bids_app = self.get_file_path("example_bad")
        self.assertRaises(
            ValidationError,
            bosh,
            ["import", "bids", "test-import.json", bids_app],
        )

    def test_upgrade_04(self):
        self.setup("import")
        fin = self.get_file_path("upgrade04.json")
        fout = self.get_file_path("upgrade05.json")
        ref_file = self.get_file_path("test-import-04-ref.json")
        ref_file_p2 = self.get_file_path("test-import-04-ref-python2.json")
        if op.isfile(fout):
            os.remove(fout)
        self.assertFalse(bosh(["import", "0.4", fout, fin]))
        result = json.loads(open(fout).read().strip())
        self.assertIn(
            result,
            [
                json.loads(open(ref_file).read().strip()),
                json.loads(open(ref_file_p2).read().strip()),
            ],
        )
        os.remove(fout)

    def test_upgrade_04_json_obj(self):
        self.setup("import")
        fin = open(self.get_file_path("upgrade04.json")).read()
        fout = self.get_file_path("upgrade05.json")
        ref_file = self.get_file_path("test-import-04-ref.json")
        ref_file_p2 = self.get_file_path("test-import-04-ref-python2.json")
        if op.isfile(fout):
            os.remove(fout)
        self.assertFalse(bosh(["import", "0.4", fout, fin]))
        result = json.loads(open(fout).read().strip())
        self.assertIn(
            result,
            [
                json.loads(open(ref_file).read().strip()),
                json.loads(open(ref_file_p2).read().strip()),
            ],
        )
        os.remove(fout)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_import_cwl_valid(self):
        self.setup("import/cwl/")
        ex_dir = self.get_file_path("")
        # These ones are supposed to crash
        bad_dirs = [
            "1st-workflow",  # workflow
            "record",  # complex type
            "array-inputs",  # input bindings specific to array element
            "expression",  # Javascript expression
            "nestedworkflows",  # workflow
        ]
        for d in os.listdir(ex_dir):
            if d == "README.md":
                continue
            cwl_descriptor = op.abspath(opj(ex_dir, d, d + ".cwl"))
            cwl_invocation = op.abspath(opj(ex_dir, d, d + ".yml"))
            assert os.path.isfile(cwl_descriptor)
            out_desc = "./test_temp/cwl_out.json"
            out_inv = "./test_temp/cwl_inv_out.json"
            run = False
            if os.path.isfile(cwl_invocation):
                args = [
                    "import",
                    "cwl",
                    out_desc,
                    cwl_descriptor,
                    "-i",
                    cwl_invocation,
                    "-o",
                    out_inv,
                ]
                run = True
            else:
                args = ["import", "cwl", out_desc, cwl_descriptor]
            if d in bad_dirs:
                with pytest.raises(ImportError):
                    bosh(args)
            else:
                self.assertFalse(bosh(args), cwl_descriptor)
                if run:
                    # write files required by cwl tools
                    with open("./test_temp/hello.js", "w") as f:
                        f.write("'hello'")
                    with open("./test_temp/goodbye.txt", "w") as f:
                        f.write("goodbye")
                    # closing required for Python 2.6...
                    with tarfile.open("./test_temp/hello.tar", "w") as tar:
                        tar.add("./test_temp/goodbye.txt")
                    ret = boutiques.execute(
                        "launch", out_desc, out_inv, "--skip-data-collection"
                    )
                    self.assertFalse(ret.exit_code, cwl_descriptor)

    @pytest.mark.xfail(reason="Travis to GH action transition")
    @pytest.mark.usefixtures("skip_if_no_apptainer")
    def test_docopt_import_valid(self):
        self.setup("import/docopt/")
        pydocopt_input = self.get_file_path("docopt_script_valid.py")
        output_descriptor = self.get_file_path("test_valid_output.json")

        import_args = ["import", "docopt", output_descriptor, pydocopt_input]
        bosh(import_args)

        test_invocation = self.get_file_path("valid_invoc_mutex.json")
        launch_args = ["exec", "launch", output_descriptor, test_invocation]
        bosh(launch_args)

        os.remove(output_descriptor)

    @pytest.mark.xfail(reason="Travis to GH action transition")
    @pytest.mark.usefixtures("skip_if_no_apptainer")
    def test_docopt_import_valid_options(self):
        self.setup("import/docopt/")
        pydocopt_input = self.get_file_path("docopt_script_options.py")
        output_descriptor = self.get_file_path("test_options_output.json")

        import_args = ["import", "docopt", output_descriptor, pydocopt_input]
        bosh(import_args)

        test_invocation = self.get_file_path("test_options_invocation.json")
        launch_args = ["exec", "launch", output_descriptor, test_invocation]
        bosh(launch_args)

        os.remove(output_descriptor)

    def test_docopt_import_invalid(self):
        self.setup("import/docopt/")
        pydocopt_input = self.get_file_path("docopt_script_invalid.py")
        output_descriptor = self.get_file_path("foobar.json")

        args = ["import", "docopt", output_descriptor, pydocopt_input]

        with pytest.raises(ImportError, match="Invalid docopt script"):
            bosh(args)
            self.fail(
                "Did not raise ImportError or"
                + " message did not match Invalid docopt script"
            )

        if op.isfile(output_descriptor):
            self.fail("Output file should not exist")

    def test_docopt_nf(self):
        self.setup("import/docopt/")
        pydocopt_input = self.get_file_path("naval_fate.py")
        output_descriptor = self.get_file_path("naval_fate_descriptor.json")

        import_args = ["import", "docopt", output_descriptor, pydocopt_input]
        bosh(import_args)

        test_invocation = self.get_file_path("nf_invoc_new.json")
        launch_args = [
            "exec",
            "simulate",
            output_descriptor,
            "-i",
            test_invocation,
        ]
        bosh(launch_args)

        test_invocation = self.get_file_path("nf_invoc_move.json")
        launch_args = [
            "exec",
            "simulate",
            output_descriptor,
            "-i",
            test_invocation,
        ]
        bosh(launch_args)

        test_invocation = self.get_file_path("nf_invoc_shoot.json")
        launch_args = [
            "exec",
            "simulate",
            output_descriptor,
            "-i",
            test_invocation,
        ]
        bosh(launch_args)

        test_invocation = self.get_file_path("nf_invoc_mine.json")
        launch_args = [
            "exec",
            "simulate",
            output_descriptor,
            "-i",
            test_invocation,
        ]
        bosh(launch_args)

        test_invocation = self.get_file_path("nf_invoc_help.json")
        launch_args = [
            "exec",
            "simulate",
            output_descriptor,
            "-i",
            test_invocation,
        ]
        bosh(launch_args)

        os.remove(output_descriptor)

    def test_import_json_config(self):
        self.setup("import/config/")
        expected_desc = loadJson(self.get_file_path("json_config_desc.json"))
        config = self.get_file_path("configuration.json")
        output_descriptor = self.get_file_path("output.json")

        import_args = ["import", "config", output_descriptor, config]
        bosh(import_args)
        result_desc = loadJson(output_descriptor)

        if op.exists(output_descriptor):
            os.remove(output_descriptor)
        self.assertEqual(expected_desc, result_desc)

        # Groups are needed in template but causes tests to fail
        del result_desc["groups"]
        # Tests the generated descriptor by running it with a test invocation
        test_invoc = self.get_file_path("test_config_import_invoc.json")
        simulate_args = [
            "exec",
            "simulate",
            json.dumps(result_desc),
            "-i",
            test_invoc,
        ]

        expected_cml = "tool config.json"
        result_cml = bosh(simulate_args).shell_command

        with open(self.get_file_path("expected_config.json")) as c:
            expect_sim_out = c.readlines()
        if op.exists(result_desc["output-files"][0]["path-template"]):
            with open(result_desc["output-files"][0]["path-template"]) as r:
                result_sim_out = r.readlines()
            os.remove(result_desc["output-files"][0]["path-template"])

        # Validate by comparing generated command-line output
        # Validate by comparing generated simulated config file
        self.assertEqual(result_cml, expected_cml)
        self.assertEqual(result_sim_out, expect_sim_out)

    def test_import_toml_config(self):
        self.setup("import/config/")
        expected_desc = loadJson(self.get_file_path("toml_config_desc.json"))
        config = self.get_file_path("configuration.toml")
        output_descriptor = self.get_file_path("output.json")

        import_args = ["import", "config", output_descriptor, config]
        bosh(import_args)
        result_desc = loadJson(output_descriptor)

        if op.exists(output_descriptor):
            os.remove(output_descriptor)
        self.assertEqual(expected_desc, result_desc)

        # Groups are needed in template but causes tests to fail
        del result_desc["groups"]
        # Tests the generated descriptor by running it with a test invocation
        test_invoc = self.get_file_path("test_config_import_invoc.json")
        simulate_args = [
            "exec",
            "simulate",
            json.dumps(result_desc),
            "-i",
            test_invoc,
        ]

        expected_cml = "tool config.toml"
        result_cml = bosh(simulate_args).shell_command

        with open(self.get_file_path("expected_config.toml")) as c:
            expect_sim_out = c.readlines()
        if op.exists(result_desc["output-files"][0]["path-template"]):
            with open(result_desc["output-files"][0]["path-template"]) as r:
                result_sim_out = r.readlines()
            os.remove(result_desc["output-files"][0]["path-template"])

        # Validate by comparing generated command-line output
        # Validate by comparing generated simulated config file
        self.assertEqual(result_cml, expected_cml)
        self.assertEqual(result_sim_out, expect_sim_out)

    def test_import_yaml_config(self):
        self.setup("import/config/")
        expected_desc = loadJson(self.get_file_path("yaml_config_desc.json"))
        config = self.get_file_path("configuration.yml")
        output_descriptor = self.get_file_path("output.json")

        import_args = ["import", "config", output_descriptor, config]
        bosh(import_args)
        result_desc = loadJson(output_descriptor)

        if op.exists(output_descriptor):
            os.remove(output_descriptor)
        self.assertEqual(expected_desc, result_desc)

        # Groups are needed in template but causes tests to fail
        del result_desc["groups"]
        # Tests the generated descriptor by running it with a test invocation
        test_invoc = self.get_file_path("test_config_import_invoc.json")
        simulate_args = [
            "exec",
            "simulate",
            json.dumps(result_desc),
            "-i",
            test_invoc,
        ]

        expected_cml = "tool config.yml"
        result_cml = bosh(simulate_args).shell_command

        with open(self.get_file_path("expected_config.yml")) as c:
            expect_sim_out = c.readlines()
        if op.exists(result_desc["output-files"][0]["path-template"]):
            with open(result_desc["output-files"][0]["path-template"]) as r:
                result_sim_out = r.readlines()
            os.remove(result_desc["output-files"][0]["path-template"])

        # Validate by comparing generated command-line output
        # Validate by comparing generated simulated config file
        self.assertEqual(result_cml, expected_cml)
        self.assertEqual(result_sim_out, expect_sim_out)



import datetime as dt
import glob
import os
import subprocess
from unittest import mock

import pytest
from boutiques_mocks import example_boutiques_tool, mock_get

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.localExec import loadJson
from boutiques.tests.BaseTest import BaseTest


def retrieve_data_record():
    data_collect_dict = {}
    cache_dir = os.path.join(os.path.expanduser("~"), ".cache", "boutiques", "data")
    if os.path.exists(cache_dir):
        cache_fls = glob.glob(cache_dir + "/*")
        if cache_fls:
            latest_file = max(cache_fls, key=os.path.getctime)
            path = os.path.join(cache_dir, latest_file)
            data_collect_dict = loadJson(path)
    return data_collect_dict


class TestDataCollection(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("data_collection")

    @pytest.fixture(autouse=True)
    def clean_up(self):
        yield
        # Clean up data collection files
        cache_dir = os.path.join(os.path.expanduser("~"), ".cache", "boutiques", "data")
        if os.path.exists(cache_dir):
            cache_fls = os.listdir(cache_dir)
            for fl in cache_fls:
                path = os.path.join(cache_dir, fl)
                st = os.stat(path)
                mtime = dt.datetime.fromtimestamp(st.st_mtime)
                if mtime > dt.datetime.now() - dt.timedelta(minutes=2):
                    os.remove(path)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_data_collection(self):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        bosh.execute(
            "launch",
            self.example1_descriptor,
            invoc,
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )
        data_collect_dict = retrieve_data_record()

        summary = data_collect_dict.get("summary")
        self.assertIsNotNone(summary)
        self.assertEqual(summary.get("name"), "Example Boutiques Tool")
        public_in = data_collect_dict.get("public-invocation")
        self.assertIsNotNone(public_in)
        self.assertEqual(public_in.get("config_num"), 4)
        self.assertEqual(public_in.get("enum_input"), "val1")
        file_input = public_in.get("file_input")
        self.assertIsNotNone(file_input)
        self.assertEqual(file_input.get("file-name"), "pyproject.toml")
        self.assertIsNotNone(file_input.get("md5sum"))
        public_out = data_collect_dict.get("public-output")
        self.assertIsNotNone(public_out)
        self.assertIn("This is stdout", public_out.get("stdout"))
        self.assertIn("This is stderr", public_out.get("stderr"))
        self.assertEqual(public_out.get("exit-code"), 0)
        self.assertEqual(public_out.get("error-message"), "")
        output_files = public_out.get("output-files")
        self.assertIsNotNone(output_files)
        self.assertIsNotNone(output_files.get("logfile"))
        self.assertIsNotNone(output_files.get("config_file"))
        logfile = output_files.get("logfile")
        self.assertEqual(logfile.get("file-name"), "log-4-coin;plop.txt")

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_skip_collection(self):
        # Register original size of cache directory for comparison
        cache_dir = os.path.join(os.path.expanduser("~"), ".cache", "boutiques", "data")
        original_size = 0
        if os.path.exists(cache_dir):
            cache_fls = os.listdir(cache_dir)
            original_size = len(cache_fls)

        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        bosh.execute(
            "launch",
            self.example1_descriptor,
            invoc,
            "--skip-data-collection",
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )

        new_size = 0
        if os.path.exists(cache_dir):
            cache_fls = os.listdir(cache_dir)
            new_size = len(cache_fls)
        self.assertEqual(new_size, original_size)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch("requests.get", return_value=mock_get())
    def test_read_doi_zenodo(self, mock_get):
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        bosh.execute(
            "launch",
            "zenodo." + str(example_boutiques_tool.id),
            invoc,
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )
        data_collect_dict = retrieve_data_record()

        summary = data_collect_dict.get("summary")
        self.assertIsNotNone(summary)
        self.assertEqual(
            summary.get("descriptor-doi"),
            "10.5281/zenodo." + str(example_boutiques_tool.id),
        )

    def test_directory_input_output(self):
        import shutil

        if os.path.exists(self.test_temp):
            shutil.rmtree(self.test_temp)
        bosh.execute(
            "launch",
            self.get_file_path("dir_input.json"),
            self.get_file_path("dir_invocation.json"),
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
        )
        data_collect_dict = retrieve_data_record()

        public_in = data_collect_dict.get("public-invocation")
        self.assertIsNotNone(public_in)
        input_dir = public_in.get("input_dir")
        self.assertIsNotNone(input_dir)
        files = input_dir.get("files")
        self.assertIsNotNone(files)
        self.assertEqual(len(files), 2)

        public_out = data_collect_dict.get("public-output")
        self.assertIsNotNone(public_out)
        output_files = public_out.get("output-files")
        self.assertIsNotNone(output_files)
        results = output_files.get("results")
        self.assertIsNotNone(results)
        files = results.get("files")
        self.assertIsNotNone(files)
        self.assertEqual(len(files), 2)

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_add_provenance(self):
        provenance = """{
            \"engine\": \"http://cbrain.ca\",
            \"dataset id\": \"1234\",
            \"cluster\": \"beluga\"
        }"""
        invoc = os.path.join(
            os.path.dirname(bfile),
            "schema",
            "examples",
            "example1",
            "invocation.json",
        )
        bosh.execute(
            "launch",
            self.example1_descriptor,
            invoc,
            "-v",
            f"{self.get_file_path('example1_mount1')}:/test_mount1",
            "-v",
            f"{self.get_file_path('example1_mount2')}:/test_mount2",
            "--provenance",
            provenance,
        )
        data_collect_dict = retrieve_data_record()

        summary = data_collect_dict.get("summary")
        self.assertIsNotNone(summary)
        self.assertEqual(summary.get("name"), "Example Boutiques Tool")
        public_in = data_collect_dict.get("public-invocation")
        self.assertIsNotNone(public_in)
        self.assertEqual(public_in.get("config_num"), 4)
        self.assertEqual(public_in.get("enum_input"), "val1")
        file_input = public_in.get("file_input")
        self.assertIsNotNone(file_input)
        self.assertEqual(file_input.get("file-name"), "pyproject.toml")
        self.assertIsNotNone(file_input.get("md5sum"))
        public_out = data_collect_dict.get("public-output")
        self.assertIsNotNone(public_out)
        self.assertIn("This is stdout", public_out.get("stdout"))
        self.assertIn("This is stderr", public_out.get("stderr"))
        self.assertEqual(public_out.get("exit-code"), 0)
        self.assertEqual(public_out.get("error-message"), "")
        output_files = public_out.get("output-files")
        self.assertIsNotNone(output_files)
        self.assertIsNotNone(output_files.get("logfile"))
        self.assertIsNotNone(output_files.get("config_file"))
        logfile = output_files.get("logfile")
        self.assertEqual(logfile.get("file-name"), "log-4-coin;plop.txt")
        provenance = data_collect_dict.get("additional-information")
        self.assertEqual(provenance.get("engine"), "http://cbrain.ca")
        self.assertEqual(provenance.get("dataset id"), "1234")
        self.assertEqual(provenance.get("cluster"), "beluga")
